#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State


class DataframeImportCodeChunk(CodeChunk):
    """
    A code chunk that generates no code, but represents when a dataframe has been
    imported by name. We want this to not be optimized out as it complicates things,
    so we make it it's own code chunk
    """

    def __init__(self, prev_state: State, display_name: str, description_comment: str):
        super().__init__(prev_state)
        self.display_name = display_name
        self.description_comment = description_comment

    def get_display_name(self) -> str:
        return self.display_name
    
    def get_description_comment(self) -> str:
        return self.description_comment

    def get_code(self) -> Tuple[List[str], List[str]]:
        return [], []


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mitosheet.mito_dash.v1.spreadsheet import Spreadsheet
from mitosheet.mito_dash.v1.mito_callback import mito_callback, activate_mito
from mitosheet.streamlit.v1 import RunnableAnalysis


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
This file contains sets of keys that let the rest of the logging infrastructure
know if this data is private or not. See the README.md file in this folder for
more details.
"""

# Params that do not need to be anonyimized
LOG_PARAMS_PUBLIC = { 'action', 'analysis_name', 'column_header_index', 'cell_editor_location', 'checklist_id', 'completed_items', 'drop', 'tour_names', 'total_number_of_tour_steps', 'created_non_empty_dataframe', 'destination_sheet_index', 'df_index_type', 'email', 'export_type', 'error', 'error_message', 'error_name', 'error_stack', 'feedback_id', 'field', 'filter_location', 'flatten_column_headers', 'format_type', 'fullscreen', 'function_name', 'graph_id', 'graph_type', 'has_headers', 'has_non_empty_filter', 'height', 'how', 'ignore_index', 'join', 'jupyterlab_theme', 'keep', 'level', 'log_event', 'message', 'move_to_deprecated_id_algorithm', 'new_column_index', 'new_dtype', 'new_graph_id', 'new_signup_step', 'new_version', 'num_args', 'num_df_args', 'num_str_args', 'num_usages', 'number_rendered_sheets', 'old_dtype', 'old_graph_id', 'old_signup_step', 'old_version', 'operator', 'paper_bgcolor', 'param_filtered', 'path_parts_length', 'plot_bgcolor', 'pro_button_location', 'questions_and_answers', 'safety_filter_turned_on_by_user', 'search_string', 'selected_element', 'sheet_index', 'sheet_index_one', 'sheet_index_two', 'sheet_indexes', 'showlegend', 'skiprows', 'sort', 'sort_direction', 'step_id_to_match', 'step_idx', 'step_type', 'steps_manager_analysis_name', 'title_font_color', 'user_agent', 'user_serch_term', 'visible', 'width', 'row_index', 'type', 'value', 'open_due_to_replay_error', 'num_invalid_imports', 'num_total_imports', 'optional_code_chunk_names', 'code_snippet_name', 'get_code_snippet_error_reason', 'completion', 'edited_completion', 'prompt', 'prompt_version', 'user_input', 'feedback', 'created_dataframe_names', 'deleted_dataframe_names', 'last_line_value', 'aiPrivacyPolicyNotAccepted', 'apiKeyNotDefined', 'feature', 'public_interface_version', 'length_of_code_with_user_edits', 'length_of_code_without_user_edits', 'js_error', 'js_error_info', 'js_taskpane_header', 'failed_log_event', 'jupyter_location'}

# Params that we want to log the length of the first element
LOG_PARAMS_LENGTH_FIRST_ELEMENT = {'optional_code'}

# Parameters that are formulas, and so need to be anonyimized in a special way
LOG_PARAMS_FORMULAS = {'new_formula', 'old_formula'}

# Parameters that are nested dicts but should not be. Note that since these can overlap
# with other titles elsewhere in different log events, we do not insist these are unique. 
# Rather, we simply only attempt to linearize them when they are actually dictonaries aka
# we can recurse on them
LOG_PARAMS_TO_LINEARIZE = {'fill_method', 'graph_creation', 'graph_preprocessing', 'graph_rendering', 'graph_styling', 'rangeslider', 'title', 'xaxis', 'yaxis', 'modified_dataframes_column_recon', 'result'}

# Almost always we keep the keys for graphs public. In some cases, we need to keep them 
# private, which we do here
LOG_PARAMS_MAP_KEYS_TO_MAKE_PRIVATE = {'values_column_ids_map'}

# We do sanity checks to make sure that there is no overlap between these sets
assert len(LOG_PARAMS_PUBLIC.intersection(LOG_PARAMS_FORMULAS)) == 0

# Keys from execution data that do not need to be anonyimized
LOG_EXECUTION_DATA_PUBLIC = {'was_series', 'num_cols_deleted', 'column_header_index', 'pandas_processing_time', 'file_delimeters', 'destination_sheet_index', 'file_encodings', 'num_cols_formatted', 'result'}

# Keys from execution data that are lists, and we just want to know the length of
LOG_EXECUTION_DATA_LENGTH_FIRST_ELEMENT = {'optional_code_that_successfully_executed'}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.


from copy import copy
from datetime import date
from typing import List, Optional, Tuple, Union

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.is_type_utils import is_datetime_dtype
from mitosheet.state import State
from mitosheet.types import (
    FC_BOOLEAN_IS_FALSE, FC_BOOLEAN_IS_TRUE, FC_DATETIME_EXACTLY,
    FC_DATETIME_GREATER, FC_DATETIME_GREATER_THAN_OR_EQUAL, FC_DATETIME_LESS,
    FC_DATETIME_LESS_THAN_OR_EQUAL, FC_DATETIME_NOT_EXACTLY, FC_EMPTY,
    FC_LEAST_FREQUENT, FC_MOST_FREQUENT, FC_NOT_EMPTY, FC_NUMBER_EXACTLY,
    FC_NUMBER_GREATER, FC_NUMBER_GREATER_THAN_OR_EQUAL, FC_NUMBER_HIGHEST,
    FC_NUMBER_LESS, FC_NUMBER_LESS_THAN_OR_EQUAL, FC_NUMBER_LOWEST,
    FC_NUMBER_NOT_EXACTLY, FC_STRING_CONTAINS, FC_STRING_DOES_NOT_CONTAIN,
    FC_STRING_ENDS_WITH, FC_STRING_EXACTLY, FC_STRING_NOT_EXACTLY,
    FC_STRING_STARTS_WITH, FC_STRING_CONTAINS_CASE_INSENSITIVE)
from mitosheet.transpiler.transpile_utils import (
    get_column_header_as_transpiled_code, get_list_as_string_without_internal_quotes)
from mitosheet.types import (ColumnHeader, ColumnID, ColumnIDWithFilterGroup,
                             Filter, FilterGroup, OperatorType)
import pandas as pd

# Dict used when a filter condition is only used by one filter
FILTER_FORMAT_STRING_DICT = {
    # SHARED CONDITIONS
    FC_EMPTY: "{df_name}[{transpiled_column_header}].isna()",
    FC_NOT_EMPTY: "{df_name}[{transpiled_column_header}].notnull()",
    FC_LEAST_FREQUENT: "{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].value_counts().index.tolist()[-{value}:])",
    FC_MOST_FREQUENT: "{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].value_counts().index.tolist()[:{value}])",
    # BOOLEAN
    FC_BOOLEAN_IS_TRUE: "{df_name}[{transpiled_column_header}] == True",
    FC_BOOLEAN_IS_FALSE: "{df_name}[{transpiled_column_header}] == False",
    # NUMBERS
    FC_NUMBER_EXACTLY: "{df_name}[{transpiled_column_header}] == {value}",
    FC_NUMBER_NOT_EXACTLY: "{df_name}[{transpiled_column_header}] != {value}",
    FC_NUMBER_GREATER: "{df_name}[{transpiled_column_header}] > {value}",
    FC_NUMBER_GREATER_THAN_OR_EQUAL: "{df_name}[{transpiled_column_header}] >= {value}",
    FC_NUMBER_LESS: "{df_name}[{transpiled_column_header}] < {value}",
    FC_NUMBER_LESS_THAN_OR_EQUAL: "{df_name}[{transpiled_column_header}] <= {value}",
    FC_NUMBER_LOWEST: '{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].nsmallest({value}, keep=\'all\'))',
    FC_NUMBER_HIGHEST: '{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].nlargest({value}, keep=\'all\'))',
    # STRINGS
    FC_STRING_CONTAINS: "{df_name}[{transpiled_column_header}].str.contains({value}, na=False, regex=False)",
    FC_STRING_DOES_NOT_CONTAIN: "~{df_name}[{transpiled_column_header}].str.contains({value}, na=False, regex=False)",
    FC_STRING_EXACTLY: "{df_name}[{transpiled_column_header}] == {value}",
    FC_STRING_NOT_EXACTLY: "{df_name}[{transpiled_column_header}] != {value}",
    FC_STRING_STARTS_WITH: "{df_name}[{transpiled_column_header}].str.startswith({value}, na=False)",
    FC_STRING_ENDS_WITH: "{df_name}[{transpiled_column_header}].str.endswith({value}, na=False)",
    FC_STRING_CONTAINS_CASE_INSENSITIVE: "{df_name}[{transpiled_column_header}].str.contains({value}, na=False, regex=False, case=False)",
    # DATES
    FC_DATETIME_EXACTLY: "{df_name}[{transpiled_column_header}] == pd.to_datetime({value})",
    FC_DATETIME_NOT_EXACTLY: "{df_name}[{transpiled_column_header}] != pd.to_datetime({value})",
    FC_DATETIME_GREATER: "{df_name}[{transpiled_column_header}] > pd.to_datetime({value})",
    FC_DATETIME_GREATER_THAN_OR_EQUAL: "{df_name}[{transpiled_column_header}] >= pd.to_datetime({value})",
    FC_DATETIME_LESS: "{df_name}[{transpiled_column_header}] < pd.to_datetime({value})",
    FC_DATETIME_LESS_THAN_OR_EQUAL: "{df_name}[{transpiled_column_header}] <= pd.to_datetime({value})",
}

# Dict used when there a specific filter condition has multiple
# filters that use it. Helps us write cleaner filter code!
FILTER_FORMAT_STRING_MULTIPLE_VALUES_DICT = {
    FC_EMPTY: {
        "Or": "{df_name}[{transpiled_column_header}].isna()",
        "And": "{df_name}[{transpiled_column_header}].isna()",
    },
    FC_NOT_EMPTY: {
        "Or": "{df_name}[{transpiled_column_header}].notnull()",
        "And": "{df_name}[{transpiled_column_header}].notnull()",
    },
    FC_LEAST_FREQUENT: {
        'Or': "{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].value_counts().index.tolist()[-max({values}):])",
        'And': "{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].value_counts().index.tolist()[-min({values}):])",
    },
    FC_MOST_FREQUENT: {
        'Or': "{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].value_counts().index.tolist()[:max({values})])",
        'And': "{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].value_counts().index.tolist()[:min({values})])"
    },
    FC_BOOLEAN_IS_TRUE: {
        "Or": "{df_name}[{transpiled_column_header}] == True",
        "And": "{df_name}[{transpiled_column_header}] == True",
    },
    FC_BOOLEAN_IS_FALSE: {
        "Or": "{df_name}[{transpiled_column_header}] == False",
        "And": "{df_name}[{transpiled_column_header}] == False",
    },
    FC_NUMBER_EXACTLY: {
        "Or": "{df_name}[{transpiled_column_header}].isin({values})",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val == n for n in {values}))",
    },
    FC_NUMBER_NOT_EXACTLY: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val != n for n in {values}))",
        "And": "~{df_name}[{transpiled_column_header}].isin({values})",
    },
    FC_NUMBER_GREATER: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val > n for n in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val > n for n in {values}))",
    },
    FC_NUMBER_GREATER_THAN_OR_EQUAL: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val >= n for n in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val >= n for n in {values}))",
    },
    FC_NUMBER_LESS: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val < n for n in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val < n for n in {values}))",
    },
    FC_NUMBER_LESS_THAN_OR_EQUAL: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val <= n for n in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val <= n for n in {values}))",
    },
    FC_NUMBER_LOWEST: {
        "Or": '{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].nsmallest(max({values}), keep=\'all\'))',
        "And": '{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].nsmallest(min({values}), keep=\'all\'))'
    }, 
    FC_NUMBER_HIGHEST: {
        "Or": '{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].nlargest(max({values}), keep=\'all\'))',
        "And": '{df_name}[{transpiled_column_header}].isin({df_name}[{transpiled_column_header}].nlargest(min({values}), keep=\'all\'))'
    },
    FC_STRING_CONTAINS: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(s in str(val) for s in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(s in str(val) for s in {values}))",
    },
    FC_STRING_DOES_NOT_CONTAIN: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(s not in str(val) for s in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(s not in str(val) for s in {values}))",
    },
    FC_STRING_EXACTLY: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val == s for s in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val == s for s in {values}))",
    },
    FC_STRING_NOT_EXACTLY: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val != s for s in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val != s for s in {values}))",
    },
    FC_STRING_STARTS_WITH: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(str(val).startswith(s) for s in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(str(val).startswith(s) for s in {values}))",
    },
    FC_STRING_ENDS_WITH: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(str(val).endswith(s) for s in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(str(val).endswith(s) for s in {values}))",
    },
    FC_STRING_CONTAINS_CASE_INSENSITIVE: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(s.upper() in str(val).upper() for s in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(s.upper() in str(val).upper() for s in {values}))",
    },
    FC_DATETIME_EXACTLY: {
        "Or": "{df_name}[{transpiled_column_header}].isin({values})",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val == d for d in {values}))",
    },
    FC_DATETIME_NOT_EXACTLY: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val != d for d in {values}))",
        "And": "~{df_name}[{transpiled_column_header}].isin({values})",
    },
    FC_DATETIME_GREATER: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val > d for d in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val > d for d in {values}))",
    },
    FC_DATETIME_GREATER_THAN_OR_EQUAL: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val >= d for d in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val >= d for d in {values}))",
    },
    FC_DATETIME_LESS: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val < d for d in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val < d for d in {values}))",
    },
    FC_DATETIME_LESS_THAN_OR_EQUAL: {
        "Or": "{df_name}[{transpiled_column_header}].apply(lambda val: any(val <= d for d in {values}))",
        "And": "{df_name}[{transpiled_column_header}].apply(lambda val: all(val <= d for d in {values}))",
    },
}

# If there are multiple conditions, we combine them together, with the
# given operator in the middle
OPERATOR_SIGNS = {"Or": "|", "And": "&"}

def get_single_filter_string(
    df_name: str, column_header: ColumnHeader, column_dtype: str, filter_: Filter
) -> str:
    """
    Transpiles a specific filter to a fitler string, to be used
    in constructing the final transpiled code
    """
    condition = filter_["condition"]
    value = filter_["value"]

    # When users are mid-way through typing a date in the UI, it might not be valid yet.
    # To prevent throwing errors, we set it to the minimum date
    if is_datetime_dtype(column_dtype):
        try: 
            pd.to_datetime(value)
        except:
            value = pd.Timestamp.min + pd.Timedelta(seconds=1)

    transpiled_column_header = get_column_header_as_transpiled_code(column_header)
    value = get_column_header_as_transpiled_code(value)

    return FILTER_FORMAT_STRING_DICT[condition].format(
        df_name=df_name, transpiled_column_header=transpiled_column_header, value=value
    )


def get_multiple_filter_string(
    df_name: str,
    column_header: ColumnHeader,
    original_operator: OperatorType,
    condition: str,
    column_dtype: str,
    filters: List[Filter],
) -> str:
    """
    Transpiles a list of filters with the same filter condition to a filter string.
    """

    # Handle dates specially by wrapping the number in a string and adding the pd.to_datetime call
    values: Union[str, List[Union[str, int, float]]]
    if is_datetime_dtype(column_dtype):
        values = [f'\'{filter["value"]}\'' for filter in filters]
        values = (
            "pd.to_datetime(" + get_list_as_string_without_internal_quotes(values) + ")"
        )
    else:
        values = [filter["value"] for filter in filters]

    transpiled_column_header = get_column_header_as_transpiled_code(column_header)

    return FILTER_FORMAT_STRING_MULTIPLE_VALUES_DICT[condition][
        original_operator
    ].format(
        df_name=df_name,
        transpiled_column_header=transpiled_column_header,
        values=values,
    )


def combine_filter_strings(
    operator: OperatorType, filter_strings: List[str], split_lines: bool = False
) -> str:
    """
    Combines the given filter strings with the passed operator, optionally
    splitting the lines at 120 characters.

    NOTE: we choose to keep groups together for readibility, and so do not
    split the lines if we are combing a group.
    """
    if len(filter_strings) == 1:
        return filter_strings[0]
    else:
        # Put parens around them
        filter_strings = [f"({fs})" for fs in filter_strings]

        filter_string = ""
        current_line_length = 0
        for i, fs in enumerate(filter_strings):
            if i != 0:
                filter_string += f" {OPERATOR_SIGNS[operator]} "
            filter_string += fs
            # We keep track of how long the lines are, and if they go over 100 characters,
            # then we split them into a new line (not if this is the last one though)
            current_line_length += len(fs)
            if (
                split_lines
                and current_line_length > 100
                and i != len(filter_strings) - 1
            ):
                filter_string += " \\\n\t"
                current_line_length = 0

        return filter_string


def create_filter_string_for_condition(
    condition: str,
    mito_filters: List[Filter],
    df_name: str,
    column_header: ColumnHeader,
    operator: OperatorType,
    column_dtype: str,
) -> str:
    """
    Returns a list of all the filter clauses for a specific filter condition in the list of passed filters
    Note: We use the nomenclature "mito_filters" here so the compiler doesn't get confused when we use the filter function
    """
    filters_with_condition = list(
        filter(
            lambda mito_filter: (mito_filter["condition"] == condition), mito_filters
        )
    )

    if len(filters_with_condition) == 0:
        return ""
    elif len(filters_with_condition) == 1:
        # Use the single filter condition
        return get_single_filter_string(
            df_name, column_header, column_dtype, filters_with_condition[0]
        )
    elif len(filters_with_condition) > 1:
        # Use the multiple filter condition
        return get_multiple_filter_string(
            df_name,
            column_header,
            operator,
            condition,
            column_dtype,
            filters_with_condition,
        )

    return ""

FAKE_COLUMN_HEADER = 'FAKE_COLUMN_HEADER'

def get_entire_filter_string_for_filters_only(df_name: str, column_header: ColumnHeader, column_dtype: str, operator: OperatorType, filters_only: List[Filter]) -> Optional[str]:

        filter_strings = []

        # We loop over the filter conditions so we avoid looping over the filters and having to ensure
        # we don't see a filter condition twice
        for filter_condition in FILTER_FORMAT_STRING_DICT.keys():
            filter_string = create_filter_string_for_condition(
                filter_condition,
                filters_only,
                df_name,
                column_header,
                operator,
                column_dtype,
            )
            if filter_string != "":
                filter_strings.append(filter_string)

        if len(filter_strings) == 0:
            return None
        elif len(filter_strings) == 1:
            return filter_strings[0]
        else:
            filter_string = combine_filter_strings(
                operator, filter_strings, split_lines=True
            )
            return filter_string


def get_entire_filter_string(state: State, sheet_index: int, operator: OperatorType, filters: List[Union[Filter, FilterGroup]], column_id: Optional[ColumnID]=None) -> Optional[str]:

        df_name = state.df_names[sheet_index]
        if column_id:
            column_header = state.column_ids.get_column_header_by_id(
                sheet_index, column_id
            )
            column_dtype = str(state.dfs[sheet_index][column_header].dtype)
        else:
            # If this filter string is for no particular column header, we use a fake column header, 
            # which allows us to change the resulting filter string to filter things other than
            # the dataframe (e.g. we want to filter a series in the conditional format). This is 
            # a somewhat ugly hack for now
            column_header = FAKE_COLUMN_HEADER
            column_dtype = 'string'

        filters_only: List[Filter] = [filter_or_group for filter_or_group in filters if "filters" not in filter_or_group] # type: ignore
        filter_groups: List[FilterGroup] = [filter_or_group for filter_or_group in filters if "filters" in filter_or_group] # type: ignore

        # Get all the filter strings
        filter_strings_with_nones = [
            get_entire_filter_string_for_filters_only(df_name, column_header, column_dtype, operator, filters_only)
        ] + [
            # NOTE: when we get to a filter group, we recurse on this function. This allows us to call this function with infinitely
            # nested filter groups, which is useful when combining code chunks
            get_entire_filter_string(state, sheet_index, filter_group["operator"], filter_group["filters"], column_id=column_id)
            for filter_group in filter_groups
        ]

        # Then, filter out all the None values
        filter_strings = [fs for fs in filter_strings_with_nones if fs is not None]

        if len(filter_strings) == 0:
            return None
        elif len(filter_strings) == 1:
            return filter_strings[0]
        else:
            filter_string = combine_filter_strings(
                operator, filter_strings, split_lines=True
            )
            return filter_string


class FilterCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, column_ids_with_filter_groups: List[ColumnIDWithFilterGroup]):
        super().__init__(prev_state)
        self.sheet_index: int = sheet_index
        self.column_ids_with_filter_groups = column_ids_with_filter_groups

        self.df_name = self.prev_state.df_names[self.sheet_index]
        column_ids = [f['column_id'] for f in self.column_ids_with_filter_groups]
        self.column_headers = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index, column_ids)

    def get_display_name(self) -> str:
        return 'Filtered'
    
    def get_description_comment(self) -> str:
        return f'Filtered {", ".join(map(str, self.column_headers))}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        all_filter_strings_with_none = [
            get_entire_filter_string(self.prev_state, self.sheet_index, f['filter']["operator"], f['filter']["filters"], f['column_id'])
            for f in self.column_ids_with_filter_groups
        ]
        all_filter_strings = [fs for fs in all_filter_strings_with_none if fs is not None]

        if len(all_filter_strings) == 0:
            return [], []
        else:
            entire_filter_string = combine_filter_strings('And', all_filter_strings, split_lines=True)
            includes_pandas = 'pd.' in entire_filter_string
            return [
                f"{self.df_name} = {self.df_name}[{entire_filter_string}]",
            ], [] if not includes_pandas else ['import pandas as pd']

    def _combine_right_with_filter_code_chunk(self, filter_code_chunk: "FilterCodeChunk") -> Optional['CodeChunk']:
        if not self.params_match(filter_code_chunk, ['sheet_index']):
            return None

        column_ids_with_filter_groups = copy(self.column_ids_with_filter_groups)
        column_ids_with_filter_groups.extend(filter_code_chunk.column_ids_with_filter_groups)

        return FilterCodeChunk(
            self.prev_state,
            self.sheet_index,
            column_ids_with_filter_groups
        )

    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, FilterCodeChunk):
            return self._combine_right_with_filter_code_chunk(other_code_chunk)

        return None

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Appends a value to a specific field, which is useful
for adding to lists in the user.json.
"""

from typing import Any

from mitosheet.telemetry.telemetry_utils import identify
from mitosheet.types import StepsManagerType
from mitosheet.user import get_user_field, set_user_field


APPEND_USER_FIELD_EVENT = 'append_user_field_update'
APPEND_USER_FIELD_PARAMS = ['field', 'value']


def execute_append_user_field_update(steps_manager: StepsManagerType, field: str, value: Any) -> None:
    """
    The function responsible for appending value to the list
    stored at field.
    """
    field_values = get_user_field(field)
    if field_values is None: 
        field_values = []
    
    # Append the new value to the list. If the value is a list itself, we 
    # extend with all the values in it
    if isinstance(value, list):
        field_values.extend(value)
    else:
        field_values.append(value)
    
    # Set the field to the new list
    set_user_field(field, field_values)

    # Identify just in case
    identify()


APPEND_USER_FIELD_UPDATE = {
    'event_type': APPEND_USER_FIELD_EVENT,
    'params': APPEND_USER_FIELD_PARAMS,
    'execute': execute_append_user_field_update
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import deepcopy
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.empty_code_chunk import EmptyCodeChunk
from mitosheet.state import State
from mitosheet.step_performers.graph_steps.graph_utils import get_graph_index_by_graph_id
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import GraphID


class GraphDuplicateStepPerformer(StepPerformer):
    """
    This steps duplicates a graph of a given graphID. 
    NOTE: WE ARE NOT USING THIS STEP IN THE CURRENT VERSION OF MITOSHEET. 
    The reason is that we separated the graph params and the graph data, 
    which caused graph duplication to not make sense as its own step. 
    Now, on the frontend we just get the graph params and send them to the backend
    as a new graph. We're keeping this around so that it doesn't break any analyses. 
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'graph_duplicate'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        old_graph_id: GraphID = get_param(params, 'old_graph_id')
        new_graph_id: GraphID = get_param(params, 'new_graph_id')

        post_state = prev_state.copy()
        graph_index = get_graph_index_by_graph_id(post_state.graph_data_array, old_graph_id)
        if (graph_index == -1):
            raise Exception(f'Graph with graph_id {old_graph_id} not found')
        graph_copy = deepcopy(post_state.graph_data_array[graph_index])
        graph_copy["graph_id"] = new_graph_id
        # We don't need to insist the the graph names are unique because they are just used in 
        # the sheet tab display. They aren't used in generated code or to identify graphs in the steps
        graph_copy["graph_tab_name"] = graph_copy["graph_tab_name"] + '_copy'        
        post_state.graph_data_array.append(graph_copy)
        
        return post_state, {
            'pandas_processing_time': 0 # No time spent on pandas, only metadata changes
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:

        # Graph steps don't add any generated code to the analysis script. 
        return [
            EmptyCodeChunk(
                prev_state, 
                'Duplicated graph',
                'Duplicated a graph',
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Union

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from mitosheet.step_performers.graph_steps.graph_utils import (BAR, BOX,
                                                               DENSITY_CONTOUR,
                                                               DENSITY_HEATMAP,
                                                               ECDF, HISTOGRAM,
                                                               LINE, SCATTER,
                                                               STRIP, VIOLIN,
                                                               get_graph_title)
from mitosheet.transpiler.transpile_utils import get_param_dict_as_code
from mitosheet.types import ColumnHeader

DO_NOT_CHANGE_PAPER_BGCOLOR_DEFAULT = '#FFFFFF'
DO_NOT_CHANGE_PLOT_BGCOLOR_DEFAULT = '#E6EBF5'
DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT = '#2F3E5D'

# The number of rows that we filter the graph to
# This must be kept in sync with GRAPH_SAFETY_FILTER_CUTOFF in GraphSidebar.tsx
GRAPH_SAFETY_FILTER_CUTOFF = 1000

# Not all of Ploty's graphs support the color parameter. Those are listed here
GRAPHS_THAT_DONT_SUPPORT_COLOR = [DENSITY_HEATMAP]

RANGE_SLIDER_CODE = """dict(
        rangeslider=dict(
            visible=True,
            thickness=.05
        )
    )"""


def safety_filter_applied(
    df: pd.DataFrame, safety_filter_turned_on_by_user: bool
) -> bool:
    """
    Helper function for determing whether the graphed dataframe
    should be filtered. It is applied if the safety_filter param is true and the
    dataframe has more than FILTERED_NUMBER_OF_ROWS rows
    """
    return (
        safety_filter_turned_on_by_user and len(df.index) > GRAPH_SAFETY_FILTER_CUTOFF
    )


def graph_filtering(
    df: pd.DataFrame, safety_filter_turned_on_by_user: bool
) -> pd.DataFrame:
    """
    Filters the dataframe to the first FILTERED_NUMBER_OF_ROWS rows, to ensure we don't crash the browser tab
    """
    if safety_filter_applied(df, safety_filter_turned_on_by_user):
        return df.head(GRAPH_SAFETY_FILTER_CUTOFF)
    else:
        return df


def graph_filtering_code(
    df_name: str, df: pd.DataFrame, safety_filter_turned_on_by_user: bool
) -> str:
    """
    Returns the code for filtering the dataframe so we don't crash the browser
    """

    if safety_filter_applied(df, safety_filter_turned_on_by_user):
        # If we do filter the graph, then return the code needed to filter the graph
        return """
# Filter the dataframe so that it does not crash the browser
{df_name}_filtered = {df_name}.head({num_rows})
""".format(
            df_name=df_name, num_rows=GRAPH_SAFETY_FILTER_CUTOFF
        )

    else:
        # If we don't filter the graph, then return an empty string
        return ""

def get_graph_creation_param_dict(
        graph_type: str,
        x_axis_column_headers: List[ColumnHeader],
        y_axis_column_headers: List[ColumnHeader],
        color_column_header: Optional[ColumnHeader],
        facet_col_column_header: Optional[ColumnHeader],
        facet_row_column_header: Optional[ColumnHeader],
        facet_col_wrap: Optional[int],
        facet_col_spacing: Optional[float],
        facet_row_spacing: Optional[float],
        points: Optional[Union[str, bool]],
        line_shape: Optional[str],
        histnorm: Optional[str],
        histfunc: Optional[str],
        nbins: Optional[int],


    ) -> Dict[str, Any]:

    # Create the parameters that we use to construct the graph
    all_params: Dict[str, Union[ColumnHeader, List[ColumnHeader], None]] = dict()

    # Plotly express requires that both the x and y parameter cannot both be lists,
    # so we need to do some casing.
    if len(x_axis_column_headers) == 1:
        # Note: In the new interface, x will always have a length of 0 or 1
        all_params["x"] = x_axis_column_headers[0]
    elif len(x_axis_column_headers) > 1:
        all_params["x"] = x_axis_column_headers

    if len(y_axis_column_headers) == 1:
        all_params["y"] = y_axis_column_headers[0]
    elif len(y_axis_column_headers) > 1:
        all_params["y"] = y_axis_column_headers

    if graph_type not in GRAPHS_THAT_DONT_SUPPORT_COLOR and color_column_header is not None:
        all_params['color'] = color_column_header

    if facet_col_column_header is not None:
        all_params['facet_col'] = facet_col_column_header

    if facet_row_column_header is not None:
        all_params['facet_row'] = facet_row_column_header

    if facet_col_wrap is not None: 
        all_params['facet_col_wrap'] = facet_col_wrap

    if facet_col_spacing is not None:
        all_params['facet_col_spacing'] = facet_col_spacing

    if facet_row_spacing is not None:
        all_params['facet_row_spacing'] = facet_row_spacing

    if points is not None:
        all_params['points'] = points

    if line_shape is not None:
        all_params['line_shape'] = line_shape

    if nbins is not None: 
        all_params['nbins'] = nbins

    if histnorm is not None:
        all_params['histnorm'] = histnorm

    if histfunc is not None:
        all_params['histfunc'] = histfunc

    return all_params

def graph_creation(
    graph_type: str,
    df: pd.DataFrame,
    x_axis_column_headers: List[ColumnHeader],
    y_axis_column_headers: List[ColumnHeader],
    color_column_header: Optional[ColumnHeader],
    facet_col_column_header: Optional[ColumnHeader],
    facet_row_column_header: Optional[ColumnHeader],
    facet_col_wrap: Optional[int],
    facet_col_spacing: Optional[float],
    facet_row_spacing: Optional[float],
    points: Optional[Union[str, bool]],
    line_shape: Optional[str],
    histnorm: Optional[str],
    histfunc: Optional[str],
    nbins: Optional[int],

) -> go.Figure:
    """
    Creates and returns the Plotly express graph figure
    """

    param_dict = get_graph_creation_param_dict(
        graph_type, 
        x_axis_column_headers, 
        y_axis_column_headers, 
        color_column_header, 
        facet_col_column_header, 
        facet_row_column_header,
        facet_col_wrap,
        facet_col_spacing,
        facet_row_spacing,
        points,
        line_shape,
        histnorm,
        histfunc,
        nbins,
    )

    # If Streamlit is installed, then we need to reset the default plotly theme.
    # Streamlit does some manipulation of the default plotly theme (https://github.com/streamlit/streamlit/blob/3d5dd61aef2bf649c5bfe42aad53cf1426beb094/lib/streamlit/elements/lib/streamlit_plotly_theme.py#L27)
    # which allows them to generate graphs that look like the rest of the app and have
    # the proper theme.
    # It turns out, this manipulation is not just when running inside Streamlit, but also
    # when importing Streamlit. So, we need to reset the default plotly theme to not display
    # graphs that are just black
    import plotly.io as pio
    if pio.templates.default == 'streamlit':
        pio.templates.default = 'plotly'

    if graph_type == BAR:
        return px.bar(df, **param_dict)
    elif graph_type == LINE:
        return px.line(df, **param_dict)
    elif graph_type == SCATTER:
        return px.scatter(df, **param_dict)
    elif graph_type == HISTOGRAM:
        return px.histogram(df, **param_dict)
    elif graph_type == DENSITY_HEATMAP:
        return px.density_heatmap(df, **param_dict)
    elif graph_type == DENSITY_CONTOUR:
        return px.density_contour(df, **param_dict)
    elif graph_type == BOX:
        return px.box(df, **param_dict)
    elif graph_type == VIOLIN:
        return px.violin(df, **param_dict)
    elif graph_type == STRIP:
        return px.strip(df, **param_dict)
    elif graph_type == ECDF:
        return px.ecdf(df, **param_dict)


def graph_creation_code(
    graph_type: str,
    df_name: str,
    x_axis_column_headers: List[ColumnHeader],
    y_axis_column_headers: List[ColumnHeader],
    color_column_header: Optional[ColumnHeader],
    facet_col_column_header: Optional[ColumnHeader],
    facet_row_column_header: Optional[ColumnHeader],
    facet_col_wrap: Optional[int],
    facet_col_spacing: Optional[float],
    facet_row_spacing: Optional[float],
    points: Optional[Union[str, bool]],
    line_shape: Optional[str],
    histnorm: Optional[str],
    histfunc: Optional[str],
    nbins: Optional[int],

) -> str:
    """
    Returns the code for creating the Plotly express graph
    """

    param_dict = get_graph_creation_param_dict(
        graph_type, 
        x_axis_column_headers, 
        y_axis_column_headers, 
        color_column_header, 
        facet_col_column_header, 
        facet_row_column_header,
        facet_col_wrap,
        facet_col_spacing,
        facet_row_spacing,
        points,
        line_shape,
        histnorm,
        histfunc,
        nbins,
    )
    param_code = get_param_dict_as_code(param_dict, as_single_line=True)

    if graph_type == BAR:
        return f"fig = px.bar({df_name}, {param_code})"
    elif graph_type == LINE:
        return f"fig = px.line({df_name}, {param_code})"
    elif graph_type == SCATTER:
        return f"fig = px.scatter({df_name}, {param_code})"
    elif graph_type == HISTOGRAM:
        return f"fig = px.histogram({df_name}, {param_code})"
    elif graph_type == DENSITY_HEATMAP:
        return f"fig = px.density_heatmap({df_name}, {param_code})"
    elif graph_type == DENSITY_CONTOUR:
        return f"fig = px.density_contour({df_name}, {param_code})"
    elif graph_type == BOX:
        return f"fig = px.box({df_name}, {param_code})"
    elif graph_type == VIOLIN:
        return f"fig = px.violin({df_name}, {param_code})"
    elif graph_type == STRIP:
        return f"fig = px.strip({df_name}, {param_code})"
    elif graph_type == ECDF:
        return f"fig = px.ecdf({df_name}, {param_code})"
    return ""

def get_graph_styling_param_dict(graph_type: str, column_headers: List[ColumnHeader], filtered: bool, graph_styling_params: Dict[str, Any]) -> Dict[str, Any]:
    """
    A param dict is a potentially nested dictonary with strings as keys with
    """
    # Create the parameters that we use to construct the graph
    all_params: Dict[str, Any] = dict()

    # Create the graph title param
    if graph_styling_params['title']['visible']:
        use_custom_title = 'title' in graph_styling_params['title']
        if use_custom_title:
            all_params['title'] = graph_styling_params['title']['title']
        else:
            all_params['title'] = get_graph_title(column_headers, [], filtered, graph_type)

        # Set the font color of the main title, if it has been changed
        title_font_color = graph_styling_params['title']['title_font_color']
        if title_font_color != DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT:
            all_params['title_font_color'] = title_font_color

    # Create the x axis param
    all_params['xaxis'] = dict()
    if graph_styling_params['xaxis']['visible']:
        # If the x axis title is visible, then either dispaly the user's custom title or Ploty's default title.
        use_custom_xaxis_title = 'title' in graph_styling_params['xaxis']
        # Only apply the xaxis title if it is set because if we set it to None, then we don't get ploty's default values
        if use_custom_xaxis_title:
            all_params['xaxis']['title'] = graph_styling_params['xaxis']['title']
        
        # Set the color of the axis, if it is not default
        xaxis_title_font_color = graph_styling_params['xaxis']['title_font_color']
        if xaxis_title_font_color != DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT:
            all_params['xaxis']['title_font_color'] = xaxis_title_font_color
    else: 
        # Plotly makes us explicitly handle setting the xaxis title and yaxis title to None
        all_params['xaxis']['title'] = None

    if 'type' in graph_styling_params['xaxis']:
        all_params['xaxis']['type'] = graph_styling_params['xaxis']['type']

    all_params['xaxis']['showgrid'] = graph_styling_params['xaxis']['showgrid']

    if 'gridwidth' in graph_styling_params['xaxis']:
        all_params['xaxis']['gridwidth'] = float(graph_styling_params['xaxis']['gridwidth'])

    # Create the range slider param
    if graph_styling_params['xaxis']['rangeslider']['visible']:
        all_params['xaxis']['rangeslider'] = dict(visible=True, thickness=0.05)

    # Create the y axis title param
    all_params['yaxis'] = dict()
    if graph_styling_params['yaxis']['visible']:
        # If the y axis title is visible, then either dispaly the user's custom title or Ploty's default title.
        use_custom_yaxis_title = 'title' in graph_styling_params['yaxis']
        # Only apply the xaxis_title if it is set because if we set it to None, then we don't get ploty's default values
        if use_custom_yaxis_title:
            all_params['yaxis']['title'] = graph_styling_params['yaxis']['title']

        # Set the color of the axis, if it is not default
        yaxis_title_font_color = graph_styling_params['yaxis']['title_font_color']
        if yaxis_title_font_color != DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT:
            all_params['yaxis']['title_font_color'] = yaxis_title_font_color
    else: 
        # Plotly makes us explicitly handle setting the xaxis_title and yaxis_title to None
        all_params['yaxis']['title'] = None

    if 'type' in graph_styling_params['yaxis']:
        all_params['yaxis']['type'] = graph_styling_params['yaxis']['type']

    all_params['yaxis']['showgrid'] = graph_styling_params['yaxis']['showgrid']

    if 'gridwidth' in graph_styling_params['yaxis']:
        all_params['yaxis']['gridwidth'] = float(graph_styling_params['yaxis']['gridwidth'])

    # Create the legend params
    if graph_styling_params['showlegend']:
        all_params['legend'] = dict()
        all_params['legend']['orientation'] = graph_styling_params['legend']['orientation']

        if 'legend' in graph_styling_params:
            if 'title' in graph_styling_params['legend'] and 'text' in graph_styling_params['legend']['title']:
                all_params['legend']['title'] = dict()
                text = graph_styling_params['legend']['title']['text']
                all_params['legend']['title']['text'] = text
            
            if 'x' in graph_styling_params['legend']:
                all_params['legend']['x'] = float(graph_styling_params['legend']['x'])

            if 'y' in graph_styling_params['legend']:
                all_params['legend']['y'] = float(graph_styling_params['legend']['y'])
    else: 
        # Only add the graph styling param if it is false, otherwise we rely on Ploty default
        all_params['showlegend'] = graph_styling_params['showlegend']
   
    # Create the barmode param
    if 'barmode' in graph_styling_params:
        all_params['barmode'] = graph_styling_params['barmode']

    if 'barnorm' in graph_styling_params:
        all_params['barnorm'] = graph_styling_params['barnorm']

    # Create the background params
    paper_bgcolor = graph_styling_params['paper_bgcolor']
    if graph_styling_params != DO_NOT_CHANGE_PAPER_BGCOLOR_DEFAULT: # NOTE: we don't need to set if it's a default
        all_params['paper_bgcolor'] = paper_bgcolor
    plot_bgcolor = graph_styling_params['plot_bgcolor']
    if plot_bgcolor != DO_NOT_CHANGE_PLOT_BGCOLOR_DEFAULT: # NOTE: we don't need to set if it's a default
        all_params['plot_bgcolor'] = plot_bgcolor

    return all_params


def graph_styling(
    fig: go.Figure, graph_type: str, column_headers: List[ColumnHeader], filtered: bool, graph_styling_params: Dict[str, Any]
) -> go.Figure:
    """
    Styles the Plotly express graph figure
    """
    param_dict = get_graph_styling_param_dict(graph_type, column_headers, filtered, graph_styling_params) 

    # Actually update the style of the graph
    fig.update_layout(
        **param_dict
    )
    return fig


def graph_styling_code(
    graph_type: 
    str, column_headers: List[ColumnHeader], 
    filtered: bool,
    graph_styling_params: Dict[str, Any]
) -> str:
    """
    Returns the code for styling the Plotly express graph
    """
    param_dict = get_graph_styling_param_dict(graph_type, column_headers, filtered, graph_styling_params) 
    params_code = get_param_dict_as_code(param_dict)
    return f"fig.update_layout({params_code})"


def get_plotly_express_graph(
    graph_type: str,
    df: pd.DataFrame,
    safety_filter_turned_on_by_user: bool,
    x_axis_column_headers: List[ColumnHeader],
    y_axis_column_headers: List[ColumnHeader],
    color_column_header: Optional[ColumnHeader],
    facet_col_column_header: Optional[ColumnHeader],
    facet_row_column_header: Optional[ColumnHeader],
    facet_col_wrap: Optional[int],
    facet_col_spacing: Optional[float],
    facet_row_spacing: Optional[float],
    points: Optional[Union[str, bool]],
    line_shape: Optional[str],
    histnorm: Optional[str],
    histfunc: Optional[str],
    nbins: Optional[int],
    graph_styling_params: Dict[str, Any],
) -> go.Figure:
    """
    Generates and returns a Plotly express graph in 3 steps
    1) filtering -- make sure that dataframe is a safe size to graph
    2) graph creation -- actually construct the graph
    3) graph styling -- style the graph
    """
    all_column_headers = x_axis_column_headers + y_axis_column_headers

    # Step 1: Filtering
    is_safety_filter_applied = safety_filter_applied(
        df, safety_filter_turned_on_by_user
    )
    df = graph_filtering(df, safety_filter_turned_on_by_user)

    # Step 2: Graph Creation
    fig = graph_creation(
        graph_type, 
        df, 
        x_axis_column_headers, 
        y_axis_column_headers, 
        color_column_header, 
        facet_col_column_header, 
        facet_row_column_header,
        facet_col_wrap,
        facet_col_spacing,
        facet_row_spacing,
        points,
        line_shape,
        histnorm,
        histfunc,
        nbins,
    )

    # Step 3: Graph Styling
    fig = graph_styling(fig, graph_type, all_column_headers, is_safety_filter_applied, graph_styling_params)

    return fig


def get_plotly_express_graph_code(
    graph_type: str,
    df: pd.DataFrame,
    safety_filter_turned_on_by_user: bool,
    x_axis_column_headers: List[ColumnHeader],
    y_axis_column_headers: List[ColumnHeader],
    color_column_header: Optional[ColumnHeader],
    facet_col_column_header: Optional[ColumnHeader],
    facet_row_column_header: Optional[ColumnHeader],
    facet_col_wrap: Optional[int],
    facet_col_spacing: Optional[float],
    facet_row_spacing: Optional[float],
    points: Optional[Union[str, bool]],
    line_shape: Optional[str],
    histnorm: Optional[str],
    histfunc: Optional[str],
    nbins: Optional[int],
    graph_styling_params: Dict[str, Any],
    df_name: str,
) -> str:
    """
    Generates the code for a Plotly express graph in 3 steps
    1) filtering -- make sure that dataframe is a safe size to graph
    2) graph creation -- actually construct the graph
    3) graph styling -- style the graph
    """

    code = []
    code.append("import plotly.express as px")

    # Step 1: Filtering
    is_safety_filter_applied = safety_filter_applied(
        df, safety_filter_turned_on_by_user
    )
    if is_safety_filter_applied:
        code.append(graph_filtering_code(df_name, df, safety_filter_turned_on_by_user))
        df_name = f"{df_name}_filtered"

    # Step 2: Graph Creation
    code.append(
        "# Construct the graph and style it. Further customize your graph by editing this code."
    )
    code.append(
        "# See Plotly Documentation for help: https://plotly.com/python/plotly-express/"
    )
    code.append(
        graph_creation_code(
            graph_type, 
            df_name, 
            x_axis_column_headers, 
            y_axis_column_headers, 
            color_column_header, 
            facet_col_column_header,
            facet_row_column_header,
            facet_col_wrap,
            facet_col_spacing,
            facet_row_spacing,
            points,
            line_shape,
            histnorm,
            histfunc,
            nbins,
        )
    )

    # Step 3: Graph Styling
    all_column_headers = x_axis_column_headers + y_axis_column_headers
    code.append(
        graph_styling_code(graph_type, all_column_headers, is_safety_filter_applied, graph_styling_params)
    )

    return "\n".join(code)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.add_column_code_chunk import \
    AddColumnCodeChunk
from mitosheet.column_headers import get_column_header_id
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param


class AddColumnStepPerformer(StepPerformer):
    """""
    A add_column step, which allows you to add a column to 
    a dataframe. 
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'add_column'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_header: str = get_param(params, 'column_header')
        column_header_index: int = get_param(params, 'column_header_index')

        # If the column_header_index is out of range, then make the new column the last column
        if column_header_index < 0 or len(prev_state.dfs[sheet_index].columns) <= column_header_index:
            column_header_index = -1

        new_column_id = get_column_header_id(column_header)

        execution_data = {
            'column_header_index': column_header_index,
            'new_column_id': new_column_id
        }

        return cls.execute_through_transpile(
            prev_state,
            params,
            execution_data,
            column_headers_to_column_ids={column_header: new_column_id}
        )

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            AddColumnCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'),    
                get_param(params, 'column_header'),
                execution_data.get('column_header_index', 0) if execution_data is not None else 0,
                execution_data.get('new_column_id', '') if execution_data is not None else '',
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from dash import Dash, callback, Input, Output, html, dcc
from mitosheet.mito_dash.v1 import Spreadsheet, mito_callback, activate_mito
import pandas as pd

app = Dash(__name__)
activate_mito(app)

app.layout = html.Div([
    html.H1("Data Changer", style={'color': 'white'}),
    dcc.Interval(id='interval1', interval=5 * 1000, n_intervals=0),
    Spreadsheet(id={'type': 'spreadsheet', 'id': 'sheet'}),
])

@callback(
    Output({'type': 'spreadsheet', 'id': 'sheet'}, 'data'),
    Input('interval1', 'n_intervals')
)
def update_spreadsheet_data(n_intervals):
    print("Refreshing new data")
    # Get new df data here
    df = pd.DataFrame({'First Data': [1, 2, 3], 'Second Data': [4, 5, 6]})
    return df.to_dict('records')

if __name__ == '__main__':
    app.run_server(debug=True)


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from collections import OrderedDict
import hashlib
import importlib
import inspect
import json
import os
import pickle
import re
from typing import Any, Dict, List, Callable, Optional, Tuple, Union

import pandas as pd

from mitosheet.mito_backend import MitoBackend
from mitosheet.selection_utils import get_selected_element
from mitosheet.types import CodeOptions, ColumnDefinitions, ConditionalFormat, DefaultEditingMode, ParamMetadata, ParamType
from mitosheet.user.utils import is_pro
from mitosheet.utils import get_new_id

CURRENT_MITO_ANALYSIS_VERSION = 1

def _get_dataframe_hash(df: pd.DataFrame) -> bytes:
    """
    Returns a hash for a pandas dataframe that is consistent across runs, notably including:
    1. The column names
    2. The values of the dataframe
    3. The index of the dataframe
    4. The order of all of these

    This is necessary due to the issues described here: https://github.com/streamlit/streamlit/issues/7086
    where streamlit default hashing is not ideal for pandas dataframes, as it misses some column header and
    reordering changes. 
    """
    try:
        return hashlib.md5(
            bytes(str(pd.util.hash_pandas_object(df.columns)), 'utf-8') +
            bytes(str(pd.util.hash_pandas_object(df)), 'utf-8')
        ).digest()
    except TypeError as e:        
        # Use pickle if pandas cannot hash the object for example if
        # it contains unhashable objects.
        return b"%s" % pickle.dumps(df, pickle.HIGHEST_PROTOCOL)

def get_dataframe_hash(df: pd.DataFrame) -> bytes:
    _PANDAS_ROWS_LARGE = 100000
    _PANDAS_SAMPLE_SIZE = 10000
    
    if len(df) >= _PANDAS_ROWS_LARGE:
        df = df.sample(n=_PANDAS_SAMPLE_SIZE, random_state=0)
    
    return _get_dataframe_hash(df)

def do_dynamic_imports(code: str) -> None:
    """
    When you get back Mito code, and you want to execute it, it requires imports defined in the global scope
    of the executing process. 

    To do this, we dynamically read in the import lines, and execute them using the importlib, and then add 
    them to the global scope.
    """

    # Extract import lines from the code_str
    import_lines = [line.strip() for line in code.split("\n") if line.strip().startswith(("from ", "import "))]

    # Dynamically import modules and attributes in the global scope
    for import_line in import_lines:
        if import_line.startswith("from"):
            match = re.match(r"from (.+) import (.+)", import_line)
            module_name = match.group(1) #type: ignore
            imported_objects = match.group(2).split(",")  #type: ignore
            
            module = importlib.import_module(module_name)
            for obj in imported_objects:
                obj = obj.strip()
                if obj == "*":
                    for attr_name in dir(module):
                        if not attr_name.startswith("_"):
                            globals()[attr_name] = getattr(module, attr_name)
                else:
                    globals()[obj] = getattr(module, obj)
        else:
            split = import_line.split(' ')
            module_name = split[1]
            alias = split[3]
            
            module = importlib.import_module(module_name)
            globals()[alias if alias else module_name] = module



def get_function_from_code_unsafe(code: str) -> Optional[Callable]:
    """
    Given a string of code, returns the first function defined in the code. Notably, to do
    this, it executes the code, and then returns the first function defined in the code. 

    As it executes the full code string, you should only use this function if you trust the
    code string -- and in our case, if the function is not called.

    If no functions are defined, returns None
    """
    functions_before = [f for f in locals().values() if callable(f)]
    exec(code)
    functions = [f for f in locals().values() if callable(f) and f not in functions_before]

    # We then find the one function that was defined inside of this module -- as the above 
    # exec likely defines all the other mitosheet functions (none of which we actaully want)
    for f in functions:
        if inspect.getmodule(f) == inspect.getmodule(get_function_from_code_unsafe):
            do_dynamic_imports(code)
            return f
        
    raise ValueError(f'No functions defined in code: {code}')

# This is the class that is returned when the user sets return_type='analysis'
# It contains data that could be relevant to the streamlit developer, and is 
# used for replaying analyses. 
class RunnableAnalysis:
    def __init__(
            self,
            code: str,
            code_options: Optional[CodeOptions],
            fully_parameterized_function: str,
            param_metadata: List[ParamMetadata],
            mito_analysis_version: int=CURRENT_MITO_ANALYSIS_VERSION
        ):
        self.__code = code
        self.__code_options = code_options
        self.__fully_parameterized_function = fully_parameterized_function
        self.__param_metadata = param_metadata
        self.mito_analysis_version = mito_analysis_version
        
    def get_param_metadata(self, param_type: Optional[ParamType]=None) -> List[ParamMetadata]:
        if param_type is None:
            return self.__param_metadata
        if param_type not in ['import', 'export']:
            raise TypeError('Invalid args passed to get_param_metadata. Type must be "import" or "export"')
        return [param for param in self.__param_metadata if param['type'] == param_type]
    
    @property
    def fully_parameterized_function(self) -> str:
        return self.__fully_parameterized_function
    
    def to_json(self) -> str:
        return json.dumps({
            'code': self.__code,
            'code_options': self.__code_options,
            'fully_parameterized_function': self.__fully_parameterized_function,
            'param_metadata': self.__param_metadata,
            'mito_analysis_version': self.mito_analysis_version
        })
    
    @staticmethod
    def from_json(json_str: str) -> 'RunnableAnalysis':
        json_dict = json.loads(json_str)
        required_keys = ['code', 'code_options', 'fully_parameterized_function', 'param_metadata']
        for key in required_keys:
            if key not in json_dict:
                raise ValueError(f'Invalid json_str passed to RunnableAnalysis.from_json. Missing key {key}.')
        return RunnableAnalysis(
            json_dict['code'],
            json_dict['code_options'],
            json_dict['fully_parameterized_function'],
            json_dict['param_metadata'],
            mito_analysis_version=json_dict['mito_analysis_version']
        )
    
    def _check_correct_args_and_kwargs(self, *args: List[Any], **kwargs: Dict[str, Any]) -> None:
        """
        Checks that all the required arguments are passed, and that
        no unexpected arguments are passed, and that no arguments
        are passed multiple times.
        """
        # Error handling for required arguments
        required_args = [param['name'] for param in self.__param_metadata if param['required']]
        for index, required_arg in enumerate(required_args):
            is_kwarg = required_arg in kwargs.keys()

            # First, check if the arg was passed in as a positional argument
            if index < len(args):
                # Check if the arg was passed in as a keyword argument as well. 
                if is_kwarg:
                    raise TypeError(f'RunnableAnalysis.run() got multiple values for argument {required_arg}')

            # If it wasn't passed as a positional argument, check if it was passed as a keyword argument
            elif not is_kwarg:
                raise TypeError(f'RunnableAnalysis.run() missing required argument {required_arg}. You passed a dataframe to this analysis, but did not pass in a value for {required_arg}.')

        # Then, check the correct kwargs were passed
        for name in kwargs:
            # Raise an error if the user passes in an unexpected argument
            if not any(param for param in self.__param_metadata if param['name'] == name):
                raise TypeError(f'RunnableAnalysis.run() got an unexpected keyword argument {name}')


    
    def run(self, *args, **kwargs):
        params = {}

        # First, set the default values for all params.
        for param in self.__param_metadata:
            params[param['name']] = param['original_value']

        self._check_correct_args_and_kwargs(*args, **kwargs)

        # Then, overwrite the default values with the user provided values
        required_args = [param['name'] for param in self.__param_metadata if param['required']]

        for index, required_arg in enumerate(required_args):
            if index < len(args):
                params[required_arg] = args[index]

        # Then, overwrite the default values with the user provided values
        for name, value in kwargs.items():
            params[name] = value

        # Then, before we call the function, we make sure that the arguments are the correct types. 
        # Notably, we want to allow users to pass in dataframes for file paths (as this is often very
        # convenient), but we need to convert them to StringIO object before calling the function
        for param_name, param_value in params.items():
            param = next(param for param in self.__param_metadata if param['name'] == param_name)
            # If the user passed in a dataframe, and the param subtype starts with file, we convert it to a StringIO object
            if param['subtype'] == 'file_name_import_csv' and isinstance(param_value, pd.DataFrame):
                from io import StringIO
                params[param_name] = StringIO(param_value.to_csv(index=False))

        return get_function_from_code_unsafe(self.__fully_parameterized_function)(**params)

try:
    import streamlit.components.v1 as components
    import streamlit as st

    parent_dir = os.path.dirname(os.path.abspath(__file__))

    mito_build_dir = os.path.join(parent_dir, "mitoBuild")
    _mito_component_func = components.declare_component("my_component", path=mito_build_dir)

    message_passer_build_dr = os.path.join(parent_dir, "messagingBuild")
    _message_passer_component_func = components.declare_component("message-passer", path=message_passer_build_dr)


    def get_session_id() -> Optional[str]:
        """
        This returns the session id for the current script run. Notably, this is different
        when the user:
        1. Refreshes the page
        2. Closes the page and reopens it
        3. Is a different user
        4. Is a different browser

        It allows us to cache the Mito backend on the session id, so that it is not
        shared across users. Notably, it clearing when refreshed is the same behavior
        that streamlit caching has by default -- e.g. the button component will reset
        when the page is refreshed.

        From the streamlit docs:
        A context object that contains data for a "script run" - that is,
        data that's scoped to a single ScriptRunner execution (and therefore also
        scoped to a single connected "session").
        """
        from streamlit.runtime.scriptrunner import get_script_run_ctx
        ctx = get_script_run_ctx()
        if ctx is None:
            return None
        return ctx.session_id

    @st.cache_resource(hash_funcs={pd.DataFrame: get_dataframe_hash})
    def _get_mito_backend(
            *args: Union[pd.DataFrame, str, None], 
            _importers: Optional[List[Callable]]=None, 
            _editors: Optional[List[Callable]]=None, 
            _sheet_functions: Optional[List[Callable]]=None, 
            _code_options: Optional[CodeOptions]=None,
            _column_definitions: Optional[List[ColumnDefinitions]]=None,
            _default_editing_mode: Optional[DefaultEditingMode]=None,
            import_folder: Optional[str]=None,
            df_names: Optional[List[str]]=None,
            session_id: Optional[str]=None,
            key: Optional[str]=None # So it caches on key
        ) -> Tuple[MitoBackend, List[Any]]: 

        mito_backend = MitoBackend(
            *args, 
            import_folder=import_folder,
            user_defined_importers=_importers, user_defined_functions=_sheet_functions, user_defined_editors=_editors,
            code_options=_code_options,
            column_definitions = _column_definitions,
            default_editing_mode=_default_editing_mode,
        )

        # Make a send function that stores the responses in a list
        responses = []
        def send(response):
            responses.append(response)
        
        mito_backend.mito_send = send

        if df_names is not None and len(df_names) > 0:
            mito_backend.receive_message(
                {
                    'event': 'update_event',
                    'id': get_new_id(),
                    'type': 'args_update',
                    'params': {
                        'args': df_names
                    },
                }
            )

        return mito_backend, responses

    def message_passer_component(key: Optional[str]=None) -> Any:
        """
        This component simply passes messages from the frontend to the backend,
        so that the backend can process them before it is rendered.
        """
        component_value = _message_passer_component_func(key=key)
        return component_value


    def spreadsheet( # type: ignore
            *args: Union[pd.DataFrame, str, None], 
            sheet_functions: Optional[List[Callable]]=None, 
            importers: Optional[List[Callable]]=None, 
            editors: Optional[List[Callable]]=None, 
            df_names: Optional[List[str]]=None,
            import_folder: Optional[str]=None,
            code_options: Optional[CodeOptions]=None,
            column_definitions: Optional[List[ColumnDefinitions]]=None,
            default_editing_mode: Optional[DefaultEditingMode]=None,
            return_type: str='default',
            height: Optional[str]=None,
            key=None
        ) -> Any:
        """
        Create a new instance of the Mito spreadsheet in a streamlit app.

        Parameters
        ----------
        args: pd.Dataframe or str or None
            The arguments to pass to the Mito spreadsheet. If a dataframe is
            passed, it will be displayed as a sheet tab. If a string is passed,
            it will be read in with a pd.read_csv call. If None is passed, it 
            will be skipped.
        sheet_functions: List[Callable]
            A list of functions that can be used in the spreadsheet. Functions
            should be capitalized.
        importers: List[Callable]
            A list of functions that can be used to import dataframes. Each
            function should return a dataframe. 
        editors: List[Callable]
            A list of functions that can be used to edit dataframes. Each function
            should have `df` as the first parameter, and then should return
            a dataframe as a result.
        df_names: List[str]
            A list of names for the dataframes passed in. If None, the dataframes
            will be named df0, df1, etc.
        key: str or None
            An key that uniquely identifies this component. This must be passed
            for now, or the component will not work. Not sure why.

        Returns
        -------
        Tuple[Dict[str, pd.DataFrame], List[str]]
            A tuple. The first element is a mapping from dataframe names to the
            final dataframes. The second element is a list of lines of code
            that were executed in the Mito spreadsheet.
        """
        session_id = get_session_id()

        mito_backend, responses = _get_mito_backend(
            *args, 
            _sheet_functions=sheet_functions,
            _importers=importers, 
            _editors=editors,
            _code_options=code_options,
            _column_definitions=column_definitions,
            _default_editing_mode=default_editing_mode,
            import_folder=import_folder,
            session_id=session_id,
            df_names=df_names, 
            key=key
        )

        # Mito widgets need new ids every time a new one is displayed. As such, if
        # the key is None, we generate a new one. Notably, we do this after getting the
        # mito_backend, so that we can cache the mito_backend on the user provided key.
        if key is None:
            key = mito_backend.analysis_name

        sheet_data_json = mito_backend.steps_manager.sheet_data_json,
        analysis_data_json = mito_backend.steps_manager.analysis_data_json,
        user_profile_json = mito_backend.get_user_profile_json()

        msg = message_passer_component(key=str(key) + 'message_passer')
        if (
            msg is not None \
            and msg['id'] not in [response['id'] for response in responses] \
            and msg['analysis_name'] == mito_backend.analysis_name
        ):
            # We receive a message if:
            # 1. It is not None
            # 2. We have not already received it on this backend
            # 3. It is for this analysis. 
            # Note that the final two conditions are to prevent messages that have been sent
            # by the message passer component from being received again. This happens because
            # when a component value is set, it is always returned by the message_passer_component
            # until a new component value is set            
            mito_backend.receive_message(msg)
            
        responses_json = json.dumps(responses)

        # NOTE: selection is Optional -- as if the user has not set the return type as selected, we don't
        # waste a component value update setting the value
        selection = _mito_component_func(
            key=key, 
            sheet_data_json=sheet_data_json, analysis_data_json=analysis_data_json, user_profile_json=user_profile_json, 
            responses_json=responses_json, id=id(mito_backend),
            height=height,
            return_type=return_type
        )

        # We return a mapping from dataframe names to dataframes
        final_state = mito_backend.steps_manager.curr_step.final_defined_state
        code = "\n".join(mito_backend.steps_manager.code())

        ordered_dict = OrderedDict()
        for df_name, df in zip(final_state.df_names, final_state.dfs):
            ordered_dict[df_name] = df

        if return_type == 'default':
            return ordered_dict, code
        elif return_type == 'selection':
            return get_selected_element(final_state.dfs, selection)
        elif return_type == 'default_list':
            return final_state.dfs, code
        elif return_type == 'dfs_dict':
            return ordered_dict
        elif return_type == 'code':
            return code
        elif return_type == 'dfs_list':
            return final_state.dfs
        elif return_type == 'function':
            if code_options is None or not code_options['as_function'] or code_options['call_function']:
                raise ValueError(f"""You must set code_options with `as_function=True` and `call_function=False` in order to return a function.""")
            
            return get_function_from_code_unsafe(code)
        elif return_type == 'analysis':
            return RunnableAnalysis(code, code_options, mito_backend.fully_parameterized_function, mito_backend.param_metadata)
        else:
            raise ValueError(f'Invalid value for return_type={return_type}. Must be "default", "default_list", "dfs", "code", "dfs_list", or "function".')

    
except ImportError:
    def spreadsheet(*args, key=None): # type: ignore
        raise RuntimeError("Couldn't import streamlit. Install streamlit with `pip install streamlit` to use the mitosheet component.")


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List

def upgrade_change_column_format_1_to_remove(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Removes the change_column_format step from the analysis since we are 
		moving to set_dataframe_format

    OLD: {
        'step_version': 1, 
        'step_type': "change_column_format",
        'params': {
            sheet_index: number,
            column_ids: List[ColumnIDs],
            format_type: Dict[str, Any]
        }
    }

    NEW: 
    """
    return later_steps


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List
from mitosheet.saved_analyses.step_upgraders.utils_column_header_to_column_id import \
    replace_headers_with_id


def upgrade_merge_1_to_merge_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Upgrades from a merge 1 step to a merge 2 step, simply
    by adding the how param.

    We just set the how param to 'lookup' 

    Old format of the step: {
        'step_version': 1, 
        'step_type': "merge", 
        'params': {
            'sheet_index_one': 0
            'merge_key_one': 'A'
            'selected_columns_one': ['A', 'B', 'C']
            'sheet_index_two': 1
            'merge_key_two': 'A'
            'selected_columns_two': ['A', 'D']
        }
    }

    New format of the step: {
        'step_version': 2, 
        'step_type': "merge",
        'params': {
            'how': 'lookup'
            'sheet_index_one': 0
            'merge_key_one': 'A'
            'selected_columns_one': ['A', 'B', 'C']
            'sheet_index_two': 1
            'merge_key_two': 'A'
            'selected_columns_two': ['A', 'D']
        }
    }
    """

    params = step['params']
    # Add the how field to the params object and set it to 'lookup'
    params['how'] = 'lookup'

    return [{
        "step_version": 2, 
        "step_type": "merge", 
        "params": params
    }] + later_steps


def upgrade_merge_2_to_3(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Moves to using column id instead of column header.

    OLD: {
        "step_version": 2, 
        "step_type": "merge", 
        "params": {
            sheet_index_one: 0,
            merge_key_one: _headers_,
            selected_columns_one: List[_headers_],
            sheet_index_two: 1,
            merge_key_two: _headers_,
            selected_columns_two: List[_headers_],
        }
    }

    NEW: {
        "step_version": 3, 
        "step_type": "merge", 
        "params": {
            sheet_index_one: 0,
            merge_key_column_id_one: _ids_,
            selected_column_ids_one: List[_ids_],
            sheet_index_two: 1,
            merge_key_column_id_two: _ids_,
            selected_column_ids_two: List[_ids_],
        }
    }
    """
    params = step['params']
    params = replace_headers_with_id(params, 'merge_key_one', 'merge_key_column_id_one')
    params = replace_headers_with_id(params, 'selected_columns_one', 'selected_column_ids_one')
    params = replace_headers_with_id(params, 'merge_key_two', 'merge_key_column_id_two')
    params = replace_headers_with_id(params, 'selected_columns_two', 'selected_column_ids_two')
    
    return [{
        "step_version": 3, 
        "step_type": "merge", 
        "params": params
    }] + later_steps


def upgrade_merge_3_to_4(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Support for multiple merge keys.

    OLD: {
        "step_version": 3, 
        "step_type": "merge", 
        "params": {
            sheet_index_one: 0,
            merge_key_column_id_one: _ids_,
            selected_column_ids_one: List[_ids_],
            sheet_index_two: 1,
            merge_key_column_id_two: _ids_,
            selected_column_ids_two: List[_ids_],
        }
    }

    NEW: {
        "step_version": 4, 
        "step_type": "merge", 
        "params": {
            sheet_index_one: 0,
            sheet_index_two: 1,
            merge_key_column_ids: [[old.merge_key_column_id_one, old.merge_key_column_id_two]],
            selected_column_ids_one: List[_ids_],
            selected_column_ids_two: List[_ids_],
        }
    }
    """
    params = step['params']
    params['merge_key_column_ids'] = [[params['merge_key_column_id_one'], params['merge_key_column_id_two']]]
    del params['merge_key_column_id_one']
    del params['merge_key_column_id_two']

    return [{
        "step_version": 4, 
        "step_type": "merge", 
        "params": params
    }] + later_steps


def upgrade_merge_4_to_5(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Support for destination_sheet_index to make edits to merges. 

    OLD: {
        "step_version": 4, 
        "step_type": "merge", 
        "params": {
            sheet_index_one: 0,
            sheet_index_two: 1,
            merge_key_column_ids: [[old.merge_key_column_id_one, old.merge_key_column_id_two]],
            selected_column_ids_one: List[_ids_],
            selected_column_ids_two: List[_ids_],
        }
    }

    NEW: {
        "step_version": 5, 
        "step_type": "merge", 
        "params": {
            sheet_index_one: 0,
            sheet_index_two: 1,
            merge_key_column_ids: [[old.merge_key_column_id_one, old.merge_key_column_id_two]],
            selected_column_ids_one: List[_ids_],
            selected_column_ids_two: List[_ids_],
            destination_sheet_index: None
        }
    }
    """
    params = step['params']
    params['destination_sheet_index'] = None

    return [{
        "step_version": 5, 
        "step_type": "merge", 
        "params": params
    }] + later_steps



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional

import pandas as pd
from mitosheet.is_type_utils import is_bool_dtype, is_datetime_dtype, is_float_dtype, is_int_dtype, is_string_dtype, is_timedelta_dtype

from mitosheet.public.v3.rolling_range import RollingRange
from mitosheet.public.v3.types.bool import cast_to_bool
from mitosheet.public.v3.types.datetime import cast_series_to_datetime, cast_to_datetime
from mitosheet.public.v3.types.float import cast_to_float
from mitosheet.public.v3.types.int import cast_to_int
from mitosheet.public.v3.types.number import cast_to_number
from mitosheet.public.v3.types.str import cast_to_string
from mitosheet.public.v3.types.timedelta import cast_to_timedelta
from mitosheet.types import PrimitiveTypeName

SERIES_CONVERSION_FUNCTIONS: Dict[PrimitiveTypeName, Optional[Callable[[pd.Series], pd.Series]]] = {
    'str': None,
    'int': None,
    'float': None,
    'number': None,
    'bool': None,
    'datetime': cast_series_to_datetime,
    'timedelta': None,
}

ELEMENT_CONVERSION_FUNCTIONS: Dict[PrimitiveTypeName, Callable[[Any], Optional[Any]]] = {
    'str': cast_to_string,
    'int': cast_to_int,
    'float': cast_to_float,
    'number': cast_to_number,
    'bool': cast_to_bool,
    'datetime': cast_to_datetime,
    'timedelta': cast_to_timedelta,
}

def is_primitive_value(value: Any) -> bool:
    return isinstance(value, str) or \
        isinstance(value, int) or \
        isinstance(value, float) or \
        isinstance(value, bool) or \
        isinstance(value, datetime) or \
        isinstance(value, timedelta)
        # TODO: nan here?

def get_primitive_type_name_from_primitive_value(primitive_value: Any) -> PrimitiveTypeName:
    if isinstance(primitive_value, str):
        return 'str'
    elif isinstance(primitive_value, int):
        return 'int'
    elif isinstance(primitive_value, float):
        return 'float'
    elif isinstance(primitive_value, bool):
        return 'bool'
    elif isinstance(primitive_value, datetime):
        return 'datetime'
    elif isinstance(primitive_value, timedelta):
        return 'timedelta'
    else:
        return 'str'
    
def get_primitive_type_name_from_series(series: pd.Series) -> PrimitiveTypeName:
    dtype = str(series.dtype)
    if is_string_dtype(dtype):
        return 'str'
    elif is_int_dtype(dtype):
        return 'int'
    elif is_float_dtype(dtype):
        return 'float'
    elif is_bool_dtype(dtype):
        return 'bool'
    elif is_datetime_dtype(dtype):
        return 'datetime'
    elif is_timedelta_dtype(dtype):
        return 'timedelta'
    else:
        return 'str'


def get_arg_cast_to_type(
        target_primitive_type_name: PrimitiveTypeName, 
        arg: Any,
        primitive_types_to_ignore: Optional[List[PrimitiveTypeName]]=None
    ) -> Any:

    if primitive_types_to_ignore is None:
        primitive_types_to_ignore = []

    series_conversion_function_without_skip = SERIES_CONVERSION_FUNCTIONS[target_primitive_type_name]
    element_conversion_function_without_skip = ELEMENT_CONVERSION_FUNCTIONS[target_primitive_type_name]

    # Wrapper that make sure we skip any primitive types we want to ignore
    def element_conversion_function(arg: Any) -> Any:
        arg_type_name = get_primitive_type_name_from_primitive_value(arg)
        if arg_type_name in primitive_types_to_ignore: # type: ignore
            return arg

        return element_conversion_function_without_skip(arg)

    if series_conversion_function_without_skip is not None:
        def series_conversion_function(arg: Any) -> Any:
            arg_type_name = get_primitive_type_name_from_series(arg)
            if arg_type_name in primitive_types_to_ignore: # type: ignore
                return arg

            return series_conversion_function_without_skip(arg) # type: ignore
    else:
        series_conversion_function = None # type: ignore

    if is_primitive_value(arg):
        return element_conversion_function(arg)
    
    elif isinstance(arg, pd.Series):
        if series_conversion_function is not None:
            return series_conversion_function(arg)

        return arg.apply(element_conversion_function)

    elif isinstance(arg, pd.DataFrame):

        if series_conversion_function is not None:
            return arg.apply(lambda c: series_conversion_function(c))

        return arg.apply(lambda c: c.apply(element_conversion_function))

    elif isinstance(arg, RollingRange):
        obj = arg.obj

        if series_conversion_function is not None:
            new_obj = obj.apply(lambda c: series_conversion_function(c))
        else:
            new_obj = obj.apply(lambda c: c.apply(element_conversion_function))
            
        return RollingRange(new_obj, arg.window, arg.offset)


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from abc import ABC, abstractmethod
from typing import Any, Dict, Collection, List, Optional, Tuple

from mitosheet.types import StepsManagerType


class PreprocessStepPerformer(ABC, object):
    """
    The abstract base class for a preprocess step, which 
    executes at the start of the analysis on the arguments
    passed to the mitosheet.sheet call
    """

    @classmethod
    @abstractmethod
    def preprocess_step_version(cls) -> int:
        """
        Returns the version of the step. Changes when the parameters
        of the step change.
        """
        pass

    @classmethod
    @abstractmethod
    def preprocess_step_type(cls) -> str:
        """
        The name of the step used internally. If you change this, you must upgrade
        the step and bump the version.
        """
        pass

    @classmethod
    @abstractmethod
    def execute(cls, args: Collection[Any]) -> Tuple[List[Any], Optional[List[str]], Optional[Dict[str, Any]]]:
        """
        Execute always returns the new list of arguments, as well as the names of these arguements, as well as execution_data
        for this preprocess step.
        """
        pass

    @classmethod
    @abstractmethod
    def transpile(cls, steps_manager: StepsManagerType, execution_data: Optional[Dict[str, Any]]) -> Tuple[List[str], List[str]]:
        """
        Returns a list of the Python code lines that corresponds to this 
        preprocess step being executed
        """
        pass



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of The Mito Enterprise license.

from typing import Any, Dict, List, Optional

import requests  # type: ignore

from mitosheet.telemetry.telemetry_utils import log
from mitosheet.transpiler.transpile_utils import NEWLINE, NEWLINE_TAB
from mitosheet.types import CodeSnippet

# Global variable used to cache the custom code snippets so that when users
# open the code snippets taskpane multiple times in the same mito instantiation, 
# we can display the custom code snippets without querying the url each time.
__cached_custom_code_snippets: Optional[List[CodeSnippet]] = None

# Helper functions for generating the proper return type
def create_error_return_obj(error_message: str) -> Dict[str, str]:
    return {
            'status': 'error',
            'error_message': error_message
    }
def create_success_return_obj(code_snippets: List[CodeSnippet]) -> Dict[str, Any]:
    return {
            'status': 'success',
            'code_snippets': code_snippets 
    }

def get_code_snippets_format_error(code_snippets: Any) -> Optional[str]:
    """
    Makes sure that the code snippets are properly formatted. Returns '' if they are properly formatted,
    and otherwise returns a helpful error message. 

    Version 1 of code snippets has the following type:
    [
            {
                    Id: str
                    Name: str
                    Description: str
                    Code: List[str]
            },
            {
                    Id: str
                    Name: str
                    Description: str
                    Code: List[str]
            },
    ]
    """
    correct_code_snippets_type = f"List[{{{NEWLINE_TAB}Id: str, {NEWLINE_TAB}Name: str, {NEWLINE_TAB}Description: str, {NEWLINE_TAB}Code: List[str]{NEWLINE}}}]"

    if not isinstance(code_snippets, list):
        return f"Custom code snippets has type {type(code_snippets)}, but should be have the format: {correct_code_snippets_type}"
    
    for code_snippet in code_snippets:
        if not isinstance(code_snippet, dict):
            return f"Custom code snippets should have the format: {correct_code_snippets_type} {NEWLINE}{NEWLINE}But this code snippet was included: {code_snippet}"

        id = code_snippet.get('Id', None)
        name = code_snippet.get('Name', None)
        description = code_snippet.get('Description', None)
        code = code_snippet.get('Code', None)

        error = False
        if id is None or not isinstance(id, str):
            error = True
        if name is None or not isinstance(name, str):
            error = True
        if description is None or not isinstance(description, str):
            error = True
        if code is None or not isinstance(code, list):
            error = True
        if not error:
            for line in code:
                if not isinstance(line, str):
                    error = True
                    break

        if error:
            return f"Custom code snippets should have the format: {correct_code_snippets_type} {NEWLINE}{NEWLINE}But this code snippet was included: {code_snippet}"

    return None

def get_custom_code_snippets(mito_config_code_snippets_url: str) -> Dict[str, Any]:
    """
    Load the code snippets from the provided url and validate that they are in the correct format
    """

    # If there are cached custom code snippets, use them
    global __cached_custom_code_snippets
    if __cached_custom_code_snippets is not None:
        return create_success_return_obj(__cached_custom_code_snippets)

    # Try to load code snippets from the URL
    try:
        response = requests.get(mito_config_code_snippets_url, verify=False)
    except Exception as e:
        error_message = f"Error accessing the code snippets data from the URL. {e}" 
        log('get_code_snippet_error', {'get_code_snippet_error_reason': error_message})
        return create_error_return_obj(error_message)

    if response.status_code != 200: 
        error_message = f"Error accessing the code snippets data from the URL. Response status code: {response.status_code}"
        log('get_code_snippet_error', {'get_code_snippet_error_reason': error_message})
        return create_error_return_obj(error_message)

    # Parse the response body into JSON 
    code_snippets = response.json()

    # Validate that the code snippets are properly formatted
    code_snippet_format_error = get_code_snippets_format_error(code_snippets)

    if code_snippet_format_error is not None:
        # Note: We do not log the code_snippet_format_error because it contains user-defined code snippets
        log('get_code_snippet_error', {'get_code_snippet_error_reason': 'code_snippet_format_error'})
        return create_error_return_obj(code_snippet_format_error)
    
    # Cache the code snippets so we don't need to request them from the url next time
    __cached_custom_code_snippets = code_snippets
    log('using_custom_code_snippets')
    return create_success_return_obj(code_snippets)
        



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
The Mito package, which contains functions for creating a Mito sheet. 

To generate a new sheet, simply run:

import mitosheet
mitosheet.sheet()

If running the mitosheet.sheet() call does not work, try restarting your JupyterLab instance, refreshing your browser, and trying again. 

If it still does not work, please email jake@sagacollab.com
"""

# Public interface we want users to rely on
from mitosheet.mito_backend import sheet
from mitosheet._version import __version__

# NOTE: We always export v1 sheet functions and types as unqualified exports, as we did
# this when we started Mito. This allows us to not break existing user analyses. We should
# never add to these -- see mitosheet/public/README.md for more info
from mitosheet.public.v1 import register_analysis
from mitosheet.public.v1.sheet_functions import *
from mitosheet.public.v1.sheet_functions.types import *
from mitosheet.public.v1.utils import flatten_column_header

# We export depricated utilities, so that users can still use them if they used to
from mitosheet.api.get_column_summary_graph import (
    filter_df_to_safe_size_external as filter_df_to_safe_size,
)
from mitosheet.step_performers.bulk_old_rename.deprecated_utils import (
    make_valid_header_external as make_valid_header,
)

# Make sure the user is initalized
from mitosheet.user import initialize_user
initialize_user()

# This function is only necessary for mitosheet3, as it is used
# in jlab3 to find the extension. It is not used in jlab2
def _jupyter_labextension_paths():
    """Called by jupyterlab to load the extension"""
    return [{"src": "labextension", "dest": "mitosheet"}]


def _jupyter_nbextension_paths():
    """Called by Jupyter Notebook Server to detect if it is a valid nbextension and
    to install the widget

    Returns
    =======
    section: The section of the Jupyter Notebook Server to change.
        Must be 'notebook' for widget extensions
    src: Source directory name to copy files from. Webpack outputs generated files
        into this directory and Jupyter Notebook copies from this directory during
        widget installation
    dest: Destination directory name to install widget files to. Jupyter Notebook copies
        from `src` directory into <jupyter path>/nbextensions/<dest> directory
        during widget installation
    require: Path to importable AMD Javascript module inside the
        <jupyter path>/nbextensions/<dest> directory
    """

    return [{
        'section': 'notebook',
        'src': 'nbextension',
        'dest': 'mitosheet',
        'require': 'mitosheet/extension'
    }]

"""
Dash Integration below. 

These are the files that are loaded by the Dash app. Notably, these need to be defined in the root __init__.py file,
as Dash will look for them there. 

The namespace stuff is a bit confusing, as well, but here's what's my current understanding is:
1. These files are loaded by Dash, and they need to be loaded by the root __init__.py file
2. The namespace in these files must be the package name, which is mitosheet
3. The _namespace in the Dash spreadsheet component must be the same as the globalName in the esbuild for dash - this is how it knows to call the right function
"""

_js_dist = [
    {'relative_package_path': 'mito_dash/v1/mitoBuild/component.js', 'namespace': 'mitosheet'}, 
    {'relative_package_path': 'mito_dash/v1/mitoBuild/component.js.map', 'namespace': 'mitosheet', 'dynamic': True}
]

_css_dist = [
    {'relative_package_path': 'mito_dash/v1/mitoBuild/component.css', 'namespace': 'mitosheet'}, 
]



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from dash import Dash, html, dcc, Output, Input
from dash.exceptions import PreventUpdate
import plotly.express as px
import pandas as pd

from mitosheet.mito_dash.v1 import Spreadsheet, mito_callback, activate_mito

df = pd.read_csv('https://raw.githubusercontent.com/plotly/datasets/master/gapminder_unfiltered.csv')

app = Dash(__name__)
activate_mito(app)

app.layout = html.Div([
    Spreadsheet(df, id={'type': 'spreadsheet', 'id': 'sheet'}, track_selection=True),
    dcc.Graph(id='graph-content'),
])

@mito_callback(
    Output('graph-content', 'figure'),
    Input({'type': 'spreadsheet', 'id': 'sheet'}, 'spreadsheet_selection')
)
def update_code(spreadsheet_selection):
    if spreadsheet_selection is None:
        raise PreventUpdate

    # Get selected rows from the original dataframe
    index = spreadsheet_selection.index
    dff = df.loc[index]

    return px.line(dff, x='year', y='lifeExp', color='continent')

if __name__ == '__main__':
    app.run(debug=True)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List


def upgrade_add_column_1_to_add_column_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Upgrades from a add column 1 step to a add column 2 step, simply
    by adding the column_header_index param

    We just set the column_header_index to -1, so that it gets added to the 
    end of the dataframe, just like all previous analyses expect.

    Old format of the step: {
        "step_version": 1, 
        "step_type": "add_column", 
        'sheet_index', 
        'column_header'
    }

    New format of the step: {
        "step_version": 2, 
        "step_type": "add_column", 
        'sheet_index', 
        'column_header', 
        'column_header_index'
    }
    """
    return [{
        "step_version": 2, 
        "step_type": "add_column", 
        'sheet_index': step['sheet_index'],
        'column_header': step['column_header'],
        'column_header_index': -1 # we set the column_header_index to -1 so that it gets added to the end
    }] + later_steps



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import ast
import os
from typing import List, Optional

def is_open_ai_credentials_available() -> bool:
    OPENAI_API_KEY = os.getenv('OPENAI_API_KEY')
    return OPENAI_API_KEY is not None

def get_import_lines_to_add(code: str) -> List[str]:
    lines = []
    if 'np.' in code and 'import numpy as np' not in code:
        lines.append('import numpy as np')
    if 'pd.' in code and 'import pandas as pd' not in code:
        lines.append('import pandas as pd')
    return lines


def fix_up_missing_imports(code: str) -> str:
    import_lines = get_import_lines_to_add(code)

    for import_line in import_lines:
        code = f'{import_line}\n' + code
        
    return code

def get_code_string_from_last_expression(code: str) -> Optional[str]:

    ast_before = ast.parse(code)
    if len(ast_before.body) == 0:
        return None
    
    last_expression = ast_before.body[-1]

    code_lines = code.splitlines()
    # NOTE; these are 1-indexed, and we need make sure we add one if they are the same, so that 
    # we can actually get the line with our slice. Also, on earlier versions of Python, the end_lineno is
    # not defined; thus, we must access it through the attribute getter
    lineno = last_expression.lineno - 1
    end_lineno = last_expression.__dict__.get('end_lineno', None)
    if end_lineno is not None:
        end_lineno -= 1
        if end_lineno == lineno:
            end_lineno += 1
    relevant_lines = code_lines[lineno:end_lineno] 
    return "\n".join(relevant_lines)

def fix_final_dataframe_name(code: str, new_df_name: str, is_series: bool) -> str:
    
    last_expression_string = get_code_string_from_last_expression(code)

    if last_expression_string is None:
        return code
    
    if not is_series:
        # If it's a dataframe, just add the name. 
        return replace_last_instance_in_string(code, last_expression_string, f'{new_df_name} = {last_expression_string}')
    else:
        # If it's a series, create a dataframe
        return replace_last_instance_in_string(code, last_expression_string, f'series = {last_expression_string}\n{new_df_name} = pd.DataFrame(series, index=series.index)')


def replace_last_instance_in_string(string: str, to_find: str, to_replace: str) -> str:
    # We only want to replace the last instance, so we have to 
    # reverse the strings and use count 1
    reversed_string = string[::-1]
    reversed_to_find = to_find[::-1]
    reversed_to_replace = to_replace[::-1]
    return reversed_string.replace(reversed_to_find, reversed_to_replace, 1)[::-1]





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.import_steps.excel_range_import_code_chunk import \
    ExcelRangeImportCodeChunk
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ExcelRangeImport
from mitosheet.utils import get_valid_dataframe_names


class ExcelRangeImportStepPerformer(StepPerformer):
    """
    Allows you to import multiple ranges from a single excel sheet.
    """

    @classmethod
    def step_version(cls) -> int:
        return 6

    @classmethod
    def step_type(cls) -> str:
        return 'excel_range_import'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        
        range_imports: List[ExcelRangeImport] = get_param(params, 'range_imports')
        new_df_names = get_valid_dataframe_names(prev_state.df_names, list(map(lambda x: x['df_name'], range_imports)))

        execution_data = {
            'new_df_names': new_df_names
        }

        return cls.execute_through_transpile(
            prev_state,
            params,
            execution_data,
            new_dataframe_params={
                'df_source': DATAFRAME_SOURCE_IMPORTED,
                'new_df_names': new_df_names,
                'overwrite': None
            }
        )

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ExcelRangeImportCodeChunk(
                prev_state, 
                get_param(params, 'file_path'),
                get_param(params, 'sheet'),
                get_param(params, 'range_imports'),
                get_param(params, 'convert_csv_to_xlsx'),
                get_param(execution_data if execution_data is not None else {}, 'new_df_names')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        # Because this step is live updating, we need to just reset all of the dataframes
        # when the user overwrites a step
        return set()
    


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Callable, List, Optional, Tuple, TypeVar, Union
import numpy as np
from datetime import datetime, timedelta

import pandas as pd

from mitosheet.public.v3.rolling_range import RollingRange


PrimitiveType = TypeVar('PrimitiveType', bound=Union[str, float, int, bool, datetime, timedelta])

ResultType = Union[pd.Series, PrimitiveType]


def __get_default_value_if_value_is_none_or_nan(value: PrimitiveType, default_value: PrimitiveType) -> PrimitiveType:
    if value is None or (isinstance(value, float) and np.isnan(value)):
        return default_value
    
    return value


def __get_new_result_series_or_primitive_helper(
        default_value: PrimitiveType,
        previous_result: ResultType, 
        new_value: ResultType,
        get_new_result_from_primitive_values: Callable[[PrimitiveType, PrimitiveType], PrimitiveType],
        get_new_result_from_series: Callable[[pd.Series, pd.Series], pd.Series]
    ) -> ResultType:
    """
    This helper function is just responsible for combining two previous results into one new
    result. 

    It does so by:
    1. Filling NaN values with the default value.
    2. Delegating to the get_new_result_from_primitive_values and get_new_result_from_series functions
    """
    if isinstance(previous_result, pd.Series):
        if isinstance(new_value, pd.Series):
            result =  get_new_result_from_series(previous_result.fillna(default_value), new_value.fillna(default_value))
            return result
        else:
            return previous_result.apply(lambda v: get_new_result_from_primitive_values(
                __get_default_value_if_value_is_none_or_nan(v, default_value), 
                __get_default_value_if_value_is_none_or_nan(new_value, default_value)
            ))
    else:
        if isinstance(new_value, pd.Series):
            return new_value.apply(lambda v: get_new_result_from_primitive_values(
                __get_default_value_if_value_is_none_or_nan(previous_result, default_value), 
                __get_default_value_if_value_is_none_or_nan(v, default_value)
            ))
        else:
            return get_new_result_from_primitive_values(
                __get_default_value_if_value_is_none_or_nan(previous_result, default_value), 
                __get_default_value_if_value_is_none_or_nan(new_value, default_value), 
            )


def __get_new_result_series_or_primitive(
        default_value: PrimitiveType,
        previous_result: ResultType,
        arg: Union[PrimitiveType, None, pd.Series, RollingRange, pd.DataFrame], 
        get_primitive_value_from_dataframe: Callable[[pd.DataFrame], PrimitiveType],
        get_new_result_from_primitive_values: Callable[[PrimitiveType, PrimitiveType], PrimitiveType],
        get_new_result_from_series: Callable[[pd.Series, pd.Series], pd.Series]
    ) -> ResultType:
    """
    This helper function does the preprocessing for a single arg, and then combines it
    with the previous result to get the new result.
    """

    def get_new_result(_previous_result: ResultType, new_value: ResultType) -> ResultType:
        return __get_new_result_series_or_primitive_helper(
            default_value,
            _previous_result,
            new_value,
            get_new_result_from_primitive_values,
            get_new_result_from_series
        )

    if isinstance(arg, pd.DataFrame):
        reduced_df = get_primitive_value_from_dataframe(arg)
        return get_new_result(previous_result, reduced_df)

    elif isinstance(arg, RollingRange):
        new_series = arg.apply(lambda df: get_primitive_value_from_dataframe(df))
        return get_new_result(previous_result, new_series)
        
    elif isinstance(arg, pd.Series):
        return get_new_result(previous_result, arg)
        
    elif arg is not None:
        return get_new_result(previous_result, arg)
    
    # Otherwise the arg is None, so we just skip it
    return previous_result


def get_final_result_series_or_primitive(
        default_value: PrimitiveType,
        argv: Tuple[Union[PrimitiveType, None, pd.Series, RollingRange, pd.DataFrame], ...], 
        get_primitive_value_from_dataframe: Callable[[pd.DataFrame], PrimitiveType],
        get_new_result_from_primitive_values: Callable[[PrimitiveType, PrimitiveType], PrimitiveType],
        get_new_result_from_series: Callable[[pd.Series, pd.Series], pd.Series]
    ) -> ResultType:
    """
    This function is the main workhorse of many sheet functions that fit a common pattern:

    1. The start with a default result (default_value).
    2. They update the result with each arg in two steps, preprocessing the arg and then combining that with the result
    3. Preprocessing the arg:
        - For dataframe values, they turned into primitive values with get_primitive_value_from_dataframe
        - For rolling ranges, they are turned into series using repeated application of get_primitive_value_from_dataframe
    4. Combining with the previous result. We are either combining two primtiive values, a primitive value and a series, or two series
        - If combining two primitive values, we combine with get_new_result_from_primitive_values
        - If combining a primitive value and a series, we use a .apply on the series with get_new_result_from_primitive_values
        - If combining two series, then we use get_new_result_from_series

    This makes it much easier to specify many functions like AND, SUM, OR, etc. 

    Notably, to make get_new_result_from_primitive_values and get_new_result_from_series easier, we fill all NaN and None
    values in the primitive values and series results being combined with the default_value before sending them to 
    these functions in step (4).
    """

    result: ResultType = default_value

    for arg in argv:
        result = __get_new_result_series_or_primitive(
            default_value,
            result,
            arg,
            get_primitive_value_from_dataframe,
            get_new_result_from_primitive_values,
            get_new_result_from_series
        )

    return result 
    

def get_index_from_series(*args: Union[PrimitiveType, pd.Series]) -> pd.Index:
    if any(isinstance(arg, pd.Series) for arg in args):
        # Search for the first series and use its index
        return next(arg.index for arg in args if isinstance(arg, pd.Series))
    else:
        raise Exception("No series found in args")
    

def get_series_from_primitive_or_series(
        arg: Union[PrimitiveType, pd.Series], 
        index: pd.Index
    ) -> pd.Series:
    if isinstance(arg, pd.Series):
        return arg
    else:
        return pd.Series([arg] * len(index), index=index)


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import csv
import os
from typing import Optional, Tuple, Union

import openpyxl
from openpyxl import load_workbook

from mitosheet.excel_utils import (get_col_and_row_indexes_from_range,
                                   get_column_from_column_index)


def get_table_range(
        file_path: str, 
        sheet_name: Optional[str]=None,
        upper_left_value: Optional[Union[str, int, float, bool]]=None, 
        # NOTE: we can't change the order of these first three parameters, for backwards compatibility
        sheet_index: Optional[int]=None, 
        upper_left_value_starts_with: Optional[Union[str, int, float, bool]]=None,
        upper_left_value_contains: Optional[Union[str, int, float, bool]]=None,
        bottom_left_corner_consecutive_empty_cells: Optional[int]=None, # TODO: at some point, we should rename this to num_empty_cells_in_final_row
        bottom_left_consecutive_empty_cells_in_first_column: Optional[int]=None,
        bottom_left_value: Optional[Union[str, int, float, bool]]=None, 
        bottom_left_value_starts_with: Optional[Union[str, int, float, bool]]=None,
        bottom_left_value_contains: Optional[Union[str, int, float, bool]]=None,
        row_entirely_empty: Optional[bool]=None,
        cumulative_number_of_empty_rows: Optional[int]=None,
        consecutive_number_of_empty_rows: Optional[int]=None,
        num_columns: Optional[int]=None
) -> Optional[str]:
    """
    Given a string, this function will look through the excel tab sheet_name at the given
    file_path and find a range that meets the conditions expressed by it's parameters.
    """
    workbook = load_workbook(file_path)

    if sheet_name is None and sheet_index is None:
        raise ValueError('Either sheet_name or sheet_index must be defined')
    elif sheet_name is not None and sheet_index is not None:
        raise ValueError('Only one of sheet_name or sheet_index can be defined')
    elif sheet_name is not None:
        sheet = workbook[sheet_name]
    else:
        sheet = workbook.worksheets[sheet_index] # type : ignore

    # We get the last defined rows, so we don't waste time searching data we don't need
    dimension = sheet.calculate_dimension()
    (min_search_col, min_search_row), (max_search_col, max_search_row) = get_col_and_row_indexes_from_range(dimension)
    # Unfortunately, openpyxl indexes from 1, so we add one to treat everything in that range
    (min_search_col, min_search_row), (max_search_col, max_search_row) = (min_search_col + 1, min_search_row + 1), (max_search_col + 1, max_search_row + 1)

    # Loop over the columns one by one to find where this value is set
    min_found_col_index, min_found_row_index = None, None
    for col in sheet.iter_cols(min_row=min_search_row, max_row=max_search_row, min_col=min_search_col, max_col=max_search_col):
        for cell in col:
            if upper_left_value is not None and cell.value == upper_left_value:
                min_found_col_index, min_found_row_index = cell.column, cell.row
                break
            if upper_left_value_starts_with is not None and str(cell.value).startswith(str(upper_left_value_starts_with)):
                min_found_col_index, min_found_row_index = cell.column, cell.row
                break
            if upper_left_value_contains is not None and str(upper_left_value_contains) in str(cell.value):
                min_found_col_index, min_found_row_index = cell.column, cell.row
                break
        
        # As soon as we find something, stop looking
        if min_found_col_index is not None or min_found_row_index is not None:
            break

    if min_found_col_index is None or min_found_row_index is None:
        return None

    # Then we find find where the columns are defined to
    if num_columns is None:
        max_found_col_index = None
        for row in sheet.iter_rows(min_row=min_found_row_index, max_row=min_found_row_index, min_col=min_found_col_index):
            for cell in row:
                if cell.value is None:
                    max_found_col_index = cell.column - 1 # minus b/c this is one past the end
                    break
    else:
        max_found_col_index = min_found_col_index + num_columns - 1

    # Similarly, if we don't find any empty value in the defined cells, we set the max_col index
    # as the limit of the sheet
    if max_found_col_index is None:
        max_found_col_index = max_search_col

    # Then we find the max row index
    column = sheet[get_column_from_column_index(min_found_col_index - 1)] # We need to subtract 1 as we 0 index
    max_found_row_index = None

    # Check for number of empty cells conditions for rows
    if bottom_left_corner_consecutive_empty_cells is not None or row_entirely_empty is not None:
        for row in sheet.iter_rows(min_row=min_found_row_index, max_row=sheet.max_row+1, min_col=min_found_col_index, max_col=max_found_col_index):
            empty_count = sum([1 if c.value is None else 0 for c in row])
            if (bottom_left_corner_consecutive_empty_cells is not None and empty_count >= bottom_left_corner_consecutive_empty_cells) or \
                (row_entirely_empty is not None and empty_count >= len(row)):
                max_found_row_index = row[0].row - 1 # minus b/c this is one past the end
                break
            
    # Check for number of empty cells conditions for columns
    if max_found_row_index is None and bottom_left_consecutive_empty_cells_in_first_column is not None:
        empty_count = 0
        for cell in column:
            if cell.row <= min_found_row_index:
                continue

            if cell.value is None:
                empty_count += 1
            else:
                empty_count = 0

            # Check if we're at the end of the column, in which case the last cell is the max
            if cell.row == sheet.max_row:
                max_found_row_index = cell.row - empty_count # minus b/c we don't want to take the empty cells
                break

            if empty_count == bottom_left_consecutive_empty_cells_in_first_column:
                max_found_row_index = cell.row - empty_count # minus b/c we don't want to take the empty cells
                break
            

    if max_found_row_index is None and cumulative_number_of_empty_rows is not None:
        num_empty = 0
        for row in sheet.iter_rows(min_row=min_found_row_index, max_row=sheet.max_row+1, min_col=min_found_col_index, max_col=max_found_col_index):
            is_empty_row = all([c.value is None for c in row])
            if is_empty_row:
                num_empty += 1
            
            if num_empty >= cumulative_number_of_empty_rows:
                max_found_row_index = row[0].row - 1 # minus b/c this is one past the end
                break
            if row[0].row == sheet.max_row:
                max_found_row_index = row[0].row # Stop at the end as well
                break

    if max_found_row_index is None and consecutive_number_of_empty_rows is not None:
        num_empty = 0
        for row in sheet.iter_rows(min_row=min_found_row_index, max_row=sheet.max_row+1, min_col=min_found_col_index, max_col=max_found_col_index):
            is_empty_row = all([c.value is None for c in row])
            if is_empty_row:
                num_empty += 1
            else:
                num_empty = 0

            if num_empty >= consecutive_number_of_empty_rows:
                max_found_row_index = row[0].row - consecutive_number_of_empty_rows # minus b/c this is past the end, empty rows
                break
            if row[0].row == sheet.max_row:
                max_found_row_index = row[0].row # stop at the end as well
                break

    # Then check for other ending conditions
    if max_found_row_index is None:
        for cell in column:
            if cell.row <= min_found_row_index:
                continue
            
            # Stop as soon as we match the final value
            if bottom_left_value is not None and bottom_left_value == cell.value:
                max_found_row_index = cell.row
                break
            if bottom_left_value_starts_with is not None and str(cell.value).startswith(str(bottom_left_value_starts_with)):
                max_found_row_index = cell.row
                break
            if bottom_left_value_contains is not None and str(bottom_left_value_contains) in str(cell.value):
                max_found_row_index = cell.row
                break
            # NOTE: IF you add more conditions here, then add them to the condition below as well checking they are None
            # so that we can continue to handle the default case of finding the first empty cells
            if (bottom_left_value is None) and (bottom_left_value_starts_with is None) and (bottom_left_value_contains is None) \
                  and cell.value is None: 
                # NOTE: Check this condition last, as it's the final end condition, and for backwards compatibility
                # this means that the user is looking for the first empty cell. NOTE
                max_found_row_index = cell.row - 1 # minus b/c this is one past the end
                break

    # If we looped over the entire column without ending, then we set the max row index
    # as the length of the entire column
    if max_found_row_index is None:
        max_found_row_index = len(column)

    return f'{get_column_from_column_index(min_found_col_index - 1)}{min_found_row_index}:{get_column_from_column_index(max_found_col_index - 1)}{max_found_row_index}'


# We keep the old function name for backwards compatibility
get_table_range_from_upper_left_corner_value = get_table_range


def get_read_excel_params_from_range(range: str) -> Tuple[int, int, str]:
    ((start_col_index, start_row_index), (end_col_index, end_row_index)) = get_col_and_row_indexes_from_range(range)
    nrows = end_row_index - start_row_index
    usecols = get_column_from_column_index(start_col_index) + ':' + get_column_from_column_index(end_col_index)
    return start_row_index, nrows, usecols


def convert_csv_file_to_xlsx_file(csv_path: str, sheet_name: Union[str, int]) -> str:
    """Converts a CSV file to an XLSX file"""
    
    xlsx_path = os.path.splitext(csv_path)[0] + '_tmp.xlsx'

    # Loop over each row of the CSV and write it to the XLSX
    with open(csv_path, 'r') as csv_file:
        csv_reader = csv.reader(csv_file)
        
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.title = sheet_name if isinstance(sheet_name, str) else f'Sheet{sheet_name}'
        for row in csv_reader:
            ws.append(row)

        wb.save(xlsx_path)

    return xlsx_path


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
from typing import Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.types import ColumnHeader


class BulkOldRenameCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, column_header_renames_list: List[Dict[ColumnHeader, ColumnHeader]]):
        super().__init__(prev_state)
        self.column_header_renames_list = column_header_renames_list

    def get_display_name(self) -> str:
        return 'Bulk rename'
    
    def get_description_comment(self) -> str:
        return f'Renamed headers for compatibility with previous Mito versions'

    def get_code(self) -> Tuple[List[str], List[str]]:

        code = []
        for sheet_index, df_name in enumerate(self.prev_state.df_names):
            renames = self.column_header_renames_list[sheet_index]
            if len(renames) == 0:
                continue

            code.append(
                f'{df_name}.rename(columns={json.dumps(renames)}, inplace=True)'
            )

        if len(code) > 0:
            code.insert(0, '# Rename headers to make them work with Mito')

        return code, []
    
    def can_be_reordered_with(self, code_chunk: CodeChunk) -> bool:
        """
        The bulk old rename cannot be reordered, ever, under any 
        circumstances. It changes a bunch of column headers, and so 
        is just scary to reorder around.
        """
        return False


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
import os
import tarfile
from pathlib import Path
from typing import Any, Dict


def get_package_json() -> Dict[str, Any]:
    """
    This helper function is responsible for getting the package.json that is bundled
    with the javascript code bundle inside this Python package. 

    When the package is mitosheet (aka, jlab 2), then the mitosheet/labextension folder
    contains a .tgz zip file that contains the package.json. 

    When the package is mitosheet3 (aka, jlab3), then the mitosheet/labextension folder
    contains the package.json directly (as this is the new build process!).

    Thus, this function handles each case, to make sure we can get the current version 
    of the mitosheet package correctly.
    """
    lab_extension_folder = os.path.join(Path(__file__).parent, 'labextension')

    try:
        # First, try and get the jlab 3 case
        package_json_path = os.path.join(lab_extension_folder, 'package.json')
        return json.loads(open(package_json_path).read())
    except:
        # If we cannot get the package json directly, then it must be zipped (as it is
        # in jlab 2), and so we pull out the package json
        tar_file_name = list(filter(lambda x: x.endswith('tgz'), os.listdir(lab_extension_folder)))[0]
        tar_file_path = os.path.join(lab_extension_folder, tar_file_name)
        with tarfile.open(tar_file_path, 'r:gz') as t:
            package_json_file = t.extractfile('package/package.json')
            if package_json_file is None:
                raise Exception("Error: no package.json found in mitosheet/. Did you build the extension correctly.")
            package_json_str = package_json_file.read()
            return json.loads(package_json_str)

package_json = get_package_json()

__version__ = package_json['version']

package_name = package_json['name']



# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Any, List, Tuple, Union

import pandas as pd

"""
A series of helper functions that help you figure out which dtype we're dealing with. 

NOTE: these should be identical to the TS utilities in dtypes.tsx
"""


def is_bool_dtype(dtype: str) -> bool:
    return 'bool' == dtype

def is_int_dtype(dtype: str) -> bool:
    return 'int' in dtype

def is_float_dtype(dtype: str) -> bool:
    return 'float' in dtype

def is_string_dtype(dtype: str) -> bool:
    return dtype == 'object' or dtype == 'str' or dtype == 'string'

def is_datetime_dtype(dtype: str) -> bool:
    # NOTE: this should handle all different datetime columns, no matter
    # the timezone, as it checks for string inclusion
    return 'datetime' in dtype

def is_timedelta_dtype(dtype: str) -> bool:
    return 'timedelta' in dtype

def is_period_dtype(dtype: str) -> bool:
    return 'period' in dtype

def is_number_dtype(dtype: str) -> bool:
    return is_int_dtype(dtype) or is_float_dtype(dtype)

def is_none_type(value: Union[str, None]) -> bool:
    """
    Helper function for determining if a value should be treated as None
    """
    return True if value is None or str(value).lower() in ['nan', 'nat'] else False


def get_float_dt_td_period_columns(df: pd.DataFrame) -> Tuple[List[Any], List[Any], List[Any], List[Any]]:
    float_columns, date_columns, timedelta_columns, period_columns = [], [], [], []
    for column_header in df.columns:
        dtype = str(df[column_header].dtype)
        # NOTE: these functions are called frequently, so we put them in 
        # the order they are most likely to be true in, so we can short out
        if is_float_dtype(dtype):
            float_columns.append(column_header)
        elif is_datetime_dtype(dtype):
            date_columns.append(column_header)
        elif is_timedelta_dtype(dtype):
            timedelta_columns.append(column_header)
        elif is_period_dtype(dtype):
            period_columns.append(column_header)

    return float_columns, date_columns, timedelta_columns, period_columns


#!/usr/bin/env python
# coding: utf-8
# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.


# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.
from packaging.version import Version
import pandas as pd
import pytest
import json

from mitosheet.tests.decorators import pandas_post_1_only
from mitosheet.tests.test_utils import create_mito_wrapper_with_data, create_mito_wrapper


def get_value_helper(sheet_data, row_index, column_index):
    return sheet_data['data'][column_index]['columnData'][row_index]

def test_sheet_json_displays_floats_correctly():
    df = pd.DataFrame({
        'value': [1.0, 2.0, None],
    })

    mito = create_mito_wrapper(df)
    
    sheet_data = json.loads(mito.sheet_data_json)[0]
    assert get_value_helper(sheet_data, 0, 0) == 1.0
    assert get_value_helper(sheet_data, 1, 0) == 2.0
    assert get_value_helper(sheet_data, 2, 0) == 'NaN'


def test_sheet_json_displays_dates_correctly():
    df = pd.DataFrame({
        'name': ['alice','bob','charlie'],
        'date_of_birth': ['2005-10-25','2002-10-2','2001-11-14']
    })

    df['date_of_birth'] = pd.to_datetime(df['date_of_birth'])

    mito = create_mito_wrapper(df)
    
    sheet_data = json.loads(mito.mito_backend.get_shared_state_variables()['sheet_data_json'])[0]
    assert get_value_helper(sheet_data, 0, 1) == '2005-10-25 00:00:00'
    assert get_value_helper(sheet_data, 1, 1) == '2002-10-02 00:00:00'
    assert get_value_helper(sheet_data, 2, 1)== '2001-11-14 00:00:00'

@pandas_post_1_only
def test_sheet_displays_dates_with_non_standard_dtype():
    mito = create_mito_wrapper_with_data(['2016-01-31T19:29:50.000+0000', '2016-01-31T19:29:50.000+0000'])
    mito.change_column_dtype(0, ['A'], 'datetime')

    sheet_data = json.loads(mito.mito_backend.get_shared_state_variables()['sheet_data_json'])[0]
    assert get_value_helper(sheet_data, 0, 0) == '2016-01-31 19:29:50'
    assert get_value_helper(sheet_data, 1, 0) == '2016-01-31 19:29:50'


def test_sheet_json_holds_timed_deltas():
    df = pd.DataFrame({
        'dob': ['2005-10-23','2002-8-2 05:12:00','2001-11-14', None],
        'dob2': ['2004-10-23 10:15:15','2002-10-2','2001-07-14 14:15:00', '2005-07-14 14:15:00']
    })

    # Check if we're on pandas 2.0, and if so, use the mixed format
    # to parse the dates.
    if Version(pd.__version__) < Version('2.0'):
        format = None
    else:
        format = 'mixed'


    # Parse date with format YYYY-MM-DD HH:MM:SS
    df['dob'] = pd.to_datetime(df['dob'], format=format)
    df['dob2'] = pd.to_datetime(df['dob2'], format=format)

    mito = create_mito_wrapper(df)

    mito.set_formula('=dob - dob2', 0, 'time_deltas', True)
    
    sheet_data = json.loads(mito.mito_backend.get_shared_state_variables()['sheet_data_json'])[0]
    assert get_value_helper(sheet_data, 0, 2) == '364 days 13:44:45'
    assert get_value_helper(sheet_data, 1, 2) == '-61 days +05:12:00'
    assert get_value_helper(sheet_data, 2, 2) == '122 days 09:45:00'
    assert get_value_helper(sheet_data, 3, 2) == 'NaT'


def test_sheet_json_holds_timed_deltas_non_standard_dtype():
    december = pd.Series(pd.date_range("20121201", periods=4))
    january = pd.Series(pd.date_range("20130101", periods=4))
    td = january - december
    td.astype("timedelta64[s]")

    df = pd.DataFrame({
        'delta': td,
    })

    mito = create_mito_wrapper(df)
    
    sheet_data = json.loads(mito.mito_backend.get_shared_state_variables()['sheet_data_json'])[0]
    assert get_value_helper(sheet_data, 0, 0) == '31 days 00:00:00'
    assert get_value_helper(sheet_data, 1, 0) == '31 days 00:00:00'
    assert get_value_helper(sheet_data, 2, 0) == '31 days 00:00:00'
    assert get_value_helper(sheet_data, 3, 0) == '31 days 00:00:00'


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import json
import os
import pandas as pd
import pytest

from mitosheet.tests.test_utils import create_mito_wrapper
from mitosheet.tests.decorators import requires_flask
from mitosheet.mito_flask.v1.flatten_utils import (flatten_mito_backend_to_json,
                                                read_backend_state_string_to_mito_backend)

TEST_DFS = [
    (pd.DataFrame({"a": [1, 2], "b": ['abc', 'def'], "c": [True, False], "d": [1.0, 2.0]})),
]


@pytest.mark.parametrize("df", TEST_DFS)
@requires_flask
def test_convert_to_string(df):
    mito = create_mito_wrapper()

    df.to_csv("test.csv", index=False)
    mito.simple_import(["test.csv"])

    mb = mito.mito_backend
    s = flatten_mito_backend_to_json(mb)

    assert json.loads(s)["shared_state_variables"] == mb.get_shared_state_variables()

    backend_state = json.loads(s)["backend_state"]
    new_mb = read_backend_state_string_to_mito_backend(backend_state)
    assert new_mb.steps_manager.dfs[0].equals(df)

    os.remove("test.csv")

@requires_flask
def test_process_edit_convert_to_string_then_convert_back_still_processes():
    test_wrapper = create_mito_wrapper()
    df = pd.DataFrame({'A': [1, 2, 3], 'B': [1, 2, 3]})
    df.to_csv("test.csv", index=False)
    test_wrapper.simple_import(["test.csv"])
    test_wrapper.add_column(0, 'C')

    s = flatten_mito_backend_to_json(test_wrapper.mito_backend)
    new_mb = read_backend_state_string_to_mito_backend(json.loads(s)["backend_state"])

    assert new_mb.steps_manager.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [1, 2, 3], 'C': [0, 0, 0]}))

    new_test_wrapper = create_mito_wrapper(mito_backend=new_mb)
    new_test_wrapper.add_column(0, 'D')

    assert new_mb.steps_manager.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [1, 2, 3], 'C': [0, 0, 0], 'D': [0, 0, 0]}))

    os.remove("test.csv")
    


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from mitosheet.public.v1.sheet_functions import FUNCTIONS
from mitosheet.public.v1.sheet_functions import *
from mitosheet.public.v1.utils import flatten_column_header, deduplicate_column_headers
from mitosheet.public.v1.sheet_functions.types import to_int_series, to_boolean_series, to_float_series, to_timedelta_series, get_datetime_format
from mitosheet.public.v1.sheet_functions.types import *
import pandas as pd


def register_analysis(analysis_name):
    """
    A helper function that is used in the frontend
    to save which analysis is being run (so it can be replayed)
    """
    from mitosheet.telemetry.telemetry_utils import log
    log('ran_generated_code', {'analysis_name': analysis_name}) 


# Forwards compatible functions
from mitosheet.public.v2.excel_utils import get_table_range_from_upper_left_corner_value, get_read_excel_params_from_range, get_table_range, convert_csv_file_to_xlsx_file


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

import datetime
import os
import sys
from typing import Any, Dict, Tuple
from mitosheet.transpiler.transpile import transpile

from mitosheet.types import StepsManagerType
from mitosheet.utils import get_valid_python_identifier

def get_path_of_import_param_for_automation(
        automation_base_folder: str,
        original_value: str,
    ) -> str:
    """
    Given the original value of an import parameter, with an absolute path, returns the 
    path of the file in the new automation folder.
    """
    file_name = os.path.basename(original_value)
    return os.path.join(automation_base_folder, 'data', file_name)

def get_data_files_for_automation(
        steps_manager: StepsManagerType,
        automation_base_folder: str,
    ) -> Dict[str, str]:
    """
    Given the steps manager, returns a dictionary of the data files that should be uploaded
    to Github for the automation. This is a dictionary of new automation file paths to file contents.
    """

    data_files: Dict[str, str] = {}

    # First, we get the CSV and Excel files the user has importer
    import_params = [param for param in steps_manager.param_metadata if param['type'] == 'import']
    original_paths = [param['original_value'] for param in import_params if 'file' in param['subtype'] and param['original_value'] is not None]

    for original_path in original_paths:
        new_path = get_path_of_import_param_for_automation(
            automation_base_folder,
            original_path
        )
        data_files[new_path] = open(original_path).read()

    return data_files

def get_minute_and_hour_from_time(time: str) -> Tuple[int, int]:
    """
    Given a time in the format HH:MM, returns the hour and minute as integers
    making sure to handle conversion from the current runtime timezone to UTC.
    """
    hour = time.split(':')[0]
    minute = time.split(':')[1]

    # Handle the timezone here. Github actions run in UTC, so we need to convert the time to UTC
    # from the current timezone of the user
    current_timezone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
    hour_and_minute = datetime.datetime.strptime(f'{hour}:{minute}', '%H:%M').time()

    # Then, make a datetime object with the current timezone, and convert it to UTC
    hour_and_minute_datetime = datetime.datetime.combine(datetime.date.today(), hour_and_minute)

    # Then, we convert it to the current timezone, and then to UTC
    hour_and_minute_datetime = hour_and_minute_datetime.replace(tzinfo=current_timezone)
    hour_and_minute_datetime = hour_and_minute_datetime.astimezone(datetime.timezone.utc)

    hour_int = hour_and_minute_datetime.hour
    minute_int = hour_and_minute_datetime.minute

    return hour_int, minute_int

def get_cron_string_from_schedule(schedule: Dict[str, Any]) -> str:
    """
    Given a schedule, returns the cron string that should be used for the Github workflow.
    """
    if schedule['type'] == 'Every Day':
        hour, minute = get_minute_and_hour_from_time(schedule['time'])
        return f'{minute} {hour} * * *'

    elif schedule['type'] == 'Every Week':
        hour, minute = get_minute_and_hour_from_time(schedule['time'])
        day_of_week = schedule['dayOfWeek']
        return f'{minute} {hour} * * {day_of_week}'

    elif schedule['type'] == 'Every Month':
        hour, minute = get_minute_and_hour_from_time(schedule['time'])
        day_of_month = schedule['dayOfMonth']
        return f'{minute} {hour} {day_of_month} * *'

    else:
        raise Exception(f"Invalid schedule type: {schedule['type']}")
    
def get_automation_code(
        steps_manager: StepsManagerType,
        automation_name: str,
        automation_base_folder: str,
    ) -> str:
    """
    Returns the code that the Github action can run to run this automation. Notably, does a ton
    of work to make sure that the paths of the parameters are correct.

    Also has to do some work to make the output path exists, so it can write output files.
    """

    # First, we get the function name
    # TODO: improve this
    function_name = get_valid_python_identifier(automation_name, 'automation_function', 'func_')

    function_code_lines = transpile(
        steps_manager,
        add_comments=False,
        optimize=True,
        code_options_override={
            'import_custom_python_code': True,
            'as_function': True,
            'call_function': False,
            'function_name': function_name,
            'function_params': 'all',
        }
    )


    # After getting the function, we go through and actually look at all the parameters. We 
    # append the necessary code to the function to define the parameters and set their values

    # NOTE: in the future, this should be a proper code option -- but this needs more specifying and
    # probably some refactoring. There's a bit of debt in code options, as they are a monolith and kinda 
    # hard to change. I'm not even really sure what the interface for function_params should be -- but it 
    # feels like there's two things:
    # 1. What should be parameterized
    # 2. What should the called values of those parameters be. 
    # Perhaps, we want to make `call_function` allow a similar sort of mapping to `function_params` --
    # where you can say for these subtypes, give them these values? But this get challenging to specify
    # the most useful types for these, as we've sene with `function_params`. I think we probably should
    # do a couple more hacks, then take a step back, and design a _beautiful_ API. I think we're close...

    import_param_definitions = []
    export_param_definitions = []
    for param_metadata in steps_manager.param_metadata:
        # If it's an import, then the correct file path to use the automations folder
        if param_metadata['type'] == 'import' and param_metadata['original_value'] is not None:
            new_path = get_path_of_import_param_for_automation(
                automation_base_folder,
                param_metadata['original_value']
            )

            import_param_definitions.append(
                f'{param_metadata["name"]} = r"{new_path}"'
            )
            
        elif param_metadata['type'] == 'export':
            # Outputs write to {automation_base_folder}/runs/{timestamp}/{param_name}.csv
            file_extension = 'csv' if 'csv' in param_metadata['subtype'] else 'xlsx'
            new_path = f"{automation_base_folder}/runs/{{export_time}}/{param_metadata['name']}.{file_extension}"
            export_param_definitions.append(
                f'{param_metadata["name"]} = f"{new_path}"'
            )

    additional_package_imports = []
    export_folder_setup = []
    if len(export_param_definitions) > 0:
        additional_package_imports.append('import os')
        additional_package_imports.append('import sys')

        # Create a folder for the current timestamp
        export_folder_setup.append('# Create a folder for this run')
        export_folder_setup.append('export_time = sys.argv[1]')
        export_folder_path = f'{automation_base_folder}/runs/{{export_time}}'
        export_folder_setup.append(f'os.makedirs(f"{export_folder_path}")')

    # Then, we actually call the function
    function_call = f'{function_name}('
    for param_metadata in steps_manager.param_metadata:
        function_call += f'{param_metadata["name"]}, '
    if len(steps_manager.param_metadata) > 0:
        function_call = function_call[:-2]
    function_call += ')'

    # Add the import and export param definitions to the function code
    function_code_lines = [
        *additional_package_imports,
        *function_code_lines,
        '',
        *export_folder_setup,
        '',
        '# Run the automation',
        *import_param_definitions,
        *export_param_definitions,
        function_call
    ]

    return '\n'.join(function_code_lines)

    
def get_requirements_txt_file() -> str:
    """
    Gets the requirements.txt file that should be used for the automation -- freezing the current 
    dependency versions of the big dependencies.

    TODO: in the future, we can handle other dependencies used in this automation (how to detect 
    automatically?) -- are there every any others?
    """
    from mitosheet import __version__ as mitosheet_version
    if mitosheet_version == '0.3.131': # Handle development version -- which is currently hardcoded
        mitosheet_version = '0.1.528'

    from pandas import __version__ as pandas_version
    return f"""mitosheet=={mitosheet_version}
pandas=={pandas_version}
"""
    

def get_github_workflow_file(
        automation_name: str,
        automation_base_folder: str,
        schedule: Dict[str, Any]    
    ) -> str:
    """
    Returns the Github workflow file that runs this automation on the given schedule.
    """

    cron_string = get_cron_string_from_schedule(schedule)
    # Get the current Python version as a string of 3.10, for example
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"

    current_timestamp_variable = "${{ steps.date.outputs.date }}"
    
    return f"""
name: "Automation: {automation_name}"
on:
  schedule:
    - cron: {cron_string}
  workflow_dispatch: null
  pull_request:
    branches:
        - main
    
jobs:
  run_automation:
    name: "Run {automation_name}"
    runs-on: ubuntu-latest      
    steps:
      - uses: actions/checkout@v4
      - name: Set up Python {python_version}
        uses: actions/setup-python@v5
        with:
          python-version: {python_version}
      - name: Get the current timestamp
        id: date
        run: echo "::set-output name=date::$(date +'%Y-%m-%d-%H-%M-%S')"
      - name: Install dependencies
        run: |
          python -m pip install --upgrade pip
          pip install -r {automation_base_folder}/requirements.txt
      - name: Run automation
        run: python {automation_base_folder}/automation.py {current_timestamp_variable}
      - name: Upload results
        uses: actions/upload-artifact@v4
        with:
          name: results-{current_timestamp_variable}
          path: {automation_base_folder}/runs/{current_timestamp_variable}
"""

def get_readme_file(
        automation_name: str,
        automation_description: str
    ) -> str:
    return f"""
# {automation_name}

{automation_description}

This automation was generated by [Mito](https://trymito.io).
"""

def get_automation_files_for_new_automation(
        steps_manager: StepsManagerType, 
        automation_name: str,
        automation_description: str,
        automation_base_folder: str,
        schedule: Dict[str, Any]
    ) -> Dict[str, str]:
    """
    Returns a mapping of file path -> file contents that should be used to create a new automation.

    In practice, includes:
    1. The csv and excel files that the user has imported
    2. The automation.py file -- the code that they generated by editing Mito
    3. The requirements.txt file -- the dependencies that they used in their automation
    4. The Github Actions workflow file -- the file that runs the generated on a schedule
    5. The README file -- a description of the automation

    All files except for 4 should be in the automation base folder.
    """
    new_files: Dict[str, str] = {}

    # Add the data files
    data_files = get_data_files_for_automation(
        steps_manager,
        automation_base_folder,
    )
    new_files.update(data_files)

    # Then, we get these files from the parameters as the function call to the function
    new_files[f'{automation_base_folder}/automation.py'] = get_automation_code(
        steps_manager,
        automation_name,
        automation_base_folder,
    )
    
    # Setup a requirements.txt file
    new_files[f'{automation_base_folder}/requirements.txt'] = get_requirements_txt_file()

    # Then, we add the Github Actions workflow file
    safe_automation_name = get_valid_python_identifier(automation_name, 'automation', 'automation_')
    new_files[f'.github/workflows/{safe_automation_name}.yml'] = get_github_workflow_file(automation_name, automation_base_folder, schedule)

    # Finally, we add a README file
    new_files[f'{automation_base_folder}/README.md'] = get_readme_file(automation_name, automation_description)

    return new_files


# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

from typing import Any, Callable
from mitosheet.mito_dash.v1.spreadsheet import (ID_TYPE,
                                                WRONG_CALLBACK_ERROR_MESSAGE,
                                                Spreadsheet)

WRONG_CALLBACK_ERROR_MESSAGE_FIRST_LINE = WRONG_CALLBACK_ERROR_MESSAGE.split('\n')[0]

try:

    from dash.exceptions import PreventUpdate

    from dash import MATCH, Dash, Input, Output, State, callback

    def mito_callback(*args: Any, **kwargs: Any) -> Callable:
        """
        This is a replacement for the @callback decorator that improves the developer experience
        with using the Mito Spreadsheet component. 

        When using the @mito_callback decorator to subscribe to changes in the `spreadsheet_result` or
        `spreadsheet_selection` properties of a Mito Spreadsheet, you get the actual SpreadsheetResult
        object - not some JSON string you need to parse and work with.

        To see the documentation for this decorator, see the @mito_callback decorator documentation:
        https://docs.trymito.io/mito-for-dash/api-reference#callback-props-and-types
        """
        def function_wrapper(original_function):
            # TODO: do we need an @wraps
            def new_function(*_args, **_kwargs):
                new_args = list(_args)

                # TODO: do we need to handle kwargs? Probably. But we will do that in the future...

                for index, arg in enumerate(new_args):

                    # We go through all the arguments passed to the function, and:
                    # 1) find those that are Mito spreadsheet_result/spreadsheet_selection error strings
                    # 2) Replace them with the the actual spreadsheet results
                    # As this is a dramatically better Spreadsheet experience

                    if isinstance(arg, str) and arg.startswith(WRONG_CALLBACK_ERROR_MESSAGE_FIRST_LINE):
                        # Get the ID from the final line 
                        mito_id = arg.split('\n')[-2].strip()
                        session_key = arg.split('\n')[-1].strip()

                        spreadsheet = Spreadsheet.get_instance(mito_id, session_key)

                        if spreadsheet is None:
                            # TODO: use a more dash exception?
                            raise Exception(f"Could not find spreadsheet with mito_id {mito_id}")
                        
                        if 'spreadsheet_result' in arg:
                            # If the user is getting the result, give them the result
                            new_args[index] = spreadsheet.get_result()
                        else:
                            # If they are just getting the selection, just get the current selection
                            new_args[index] = spreadsheet.get_result().selection()


                return original_function(*new_args, **_kwargs)
            
            return callback(*args, **kwargs)(new_function)
        
        return function_wrapper
    
    def activate_mito(app: Dash) -> None:
        """
        This function must be called right after instatiated your Dash application, so that 
        Mito can correctly hook-up message handling for all Mito spreadsheet components.

        Correct usage
        ```
        from dash import Dash
        from mitosheet.mito_dash.v1 import activate_mito, Spreadsheet

        app = Dash(__name__)
        activate_mito(app)

        # Go on to create layout, register callbacks, etc
        ```

        If this function is not called, then the Mito Spreadsheet component will not work, and will throw
        an error if you try to use it.
        """

        @app.callback(
            Output({'type': ID_TYPE, 'id': MATCH}, 'all_json', allow_duplicate=True), 
            Output({'type': ID_TYPE, 'id': MATCH}, 'spreadsheet_result', allow_duplicate=True), 
            Input({'type': ID_TYPE, 'id': MATCH}, 'message'),
            State({'type': ID_TYPE, 'id': MATCH}, 'mito_id'),
            State({'type': ID_TYPE, 'id': MATCH}, 'session_key'),
            prevent_initial_call=True
        )
        def handle_message(msg, mito_id, session_key):
            spreadsheet = Spreadsheet.get_instance(mito_id, session_key)
            if spreadsheet is None:
                print("No spreadsheet found for id", mito_id)
                # TODO: should we print some error here
                raise PreventUpdate
            
            spreadsheet.num_messages += 1

            spreadsheet.unprocessed_messages.put(msg)
            spreadsheet.process_single_message(session_key)
            
            spreadsheet.spreadsheet_result = WRONG_CALLBACK_ERROR_MESSAGE.format(prop_name='spreadsheet_result', num_messages=spreadsheet.num_messages, id=spreadsheet.mito_id, session_key=session_key)
            return spreadsheet.get_all_json(), spreadsheet.spreadsheet_result
        

        @app.callback(
            Output({'type': ID_TYPE, 'id': MATCH}, 'all_json', allow_duplicate=True), 
            Output({'type': ID_TYPE, 'id': MATCH}, 'spreadsheet_selection', allow_duplicate=True), 
            Input({'type': ID_TYPE, 'id': MATCH}, 'index_and_selections'),
            State({'type': ID_TYPE, 'id': MATCH}, 'mito_id'),
            State({'type': ID_TYPE, 'id': MATCH}, 'session_key'),
            prevent_initial_call=True
        )
        def handle_selection_change(index_and_selections, mito_id, session_key):
            spreadsheet = Spreadsheet.get_instance(mito_id, session_key)
            if spreadsheet is None:
                # TODO: should we print some error here
                raise PreventUpdate

            spreadsheet.num_messages += 1

            spreadsheet.index_and_selections = index_and_selections
            
            spreadsheet.spreadsheet_selection = WRONG_CALLBACK_ERROR_MESSAGE.format(
                prop_name='spreadsheet_selection', 
                num_messages=spreadsheet.num_messages, 
                id=spreadsheet.mito_id,
                session_key=session_key
            )
            return spreadsheet.get_all_json(), spreadsheet.spreadsheet_selection
            
    
        @app.callback(
            Output({'type': ID_TYPE, 'id': MATCH}, 'all_json', allow_duplicate=True), 
            Output({'type': ID_TYPE, 'id': MATCH}, 'spreadsheet_result', allow_duplicate=True), 
            Output({'type': ID_TYPE, 'id': MATCH}, 'spreadsheet_selection', allow_duplicate=True), 
            Input({'type': ID_TYPE, 'id': MATCH}, 'data'),
            State({'type': ID_TYPE, 'id': MATCH}, 'mito_id'),
            State({'type': ID_TYPE, 'id': MATCH}, 'session_key'),
            prevent_initial_call=True
        )
        def handle_data_change_data(data, mito_id, session_key):
            spreadsheet = Spreadsheet.get_instance(mito_id, session_key)
            if spreadsheet is None:
                print("No spreadsheet found for id", mito_id)
                # TODO: should we print some error here
                raise PreventUpdate
                        
            spreadsheet._set_new_mito_backend(
                data,
                session_key=session_key,
                import_folder=spreadsheet.import_folder, 
                code_options=spreadsheet.code_options,
                df_names=spreadsheet.df_names,
                sheet_functions=spreadsheet.sheet_functions,
                importers=spreadsheet.importers,
                editors=spreadsheet.editors,
                theme=spreadsheet.theme
            )

            return spreadsheet.get_all_json(), spreadsheet.spreadsheet_result, spreadsheet.spreadsheet_selection
        

except ImportError:

    def mito_callback(*args, **kwargs): # type: ignore
        raise Exception("You must install dash to use the @mito_callback decorator component")
    
    def activate_mito(*args, **kwargs): # type: ignore
        raise Exception("You must install dash to use the @mito_callback decorator component")